from typing import Optional

import numpy as np
from numba import njit

@njit
def threshold_crossing_interp(
    arr_t: np.ndarray,
    arr_s: np.ndarray,
    threshold: float,
    n_est: Optional[float] = None,
    trigger_on_rising_edge: bool = True,
) -> np.ndarray:
    """A sequential threshold crossing algorithm that interpolates
        the ToA between the two samples where the signal crosses
        the threshold.

    Args:
        arr_t (np.ndarray): The array containing the time values.
        arr_s (np.ndarray): The array containing the signal voltage values
            corresponding to the time values.
        threshold (float): The threshold value.
        n_est (float, optional): The estimated number of ToAs in this signal. Defaults
            to None. This number is used to pre-allocate the array containing the
            ToAs. If this number is not provided, the array will be pre-allocated as
            the same dimension as arr_t and arr_s.
        trigger_on_rising_edge (bool, optional): Whether to trigger ToAs on the rising
            or falling edge. Defaults to True. If True, the ToA is triggered on
            the rising edge.

    Returns:
        np.ndarray: An array containing the ToAs.
    """

    # Pre-allocate the array containing the ToAs
    if n_est is None:
        arr_toa = -1 * np.ones(arr_t.shape)
    else:
        arr_toa = -1 * np.ones(n_est)
    
    # Initialise the index of the ToA array
    i_toa = 0

    # Initialise the previous sample value
    prev_sample = arr_s[0]

    # Loop through all the samples
    for i_sample in range(1, arr_s.shape[0]):
        # Get the current sample value
        curr_sample = arr_s[i_sample]

        # Check if the threshold is crossed
        if trigger_on_rising_edge:
            if (prev_sample < threshold) and (curr_sample >= threshold):
                # Interpolate the ToA
                if i_toa >= arr_toa.shape[0]:
                    raise ValueError(
                        "The number of ToAs has exceeded the estimated number of ToAs. "
                        f"n_est is currently {n_est}. You must increase it. This error occurred"
                        f" at the sample index {i_sample} out of {arr_s.shape[0]}. "
                        f"Your ToA array has therefore been filled within {round(100 * i_sample/arr_s.shape[0])} % "
                        "of the signal."
                    )
                arr_toa[i_toa] = arr_t[i_sample - 1] + (
                    arr_t[i_sample] - arr_t[i_sample - 1]
                ) * (threshold - prev_sample) / (curr_sample - prev_sample)
                i_toa += 1
        else:
            if (prev_sample > threshold) and (curr_sample <= threshold):
                if i_toa >= arr_toa.shape[0]:
                    raise ValueError(
                        "The number of ToAs has exceeded the estimated number of ToAs. "
                        f"n_est is currently {n_est}. You must increase it. This error occurred"
                        f" at the sample index {i_sample} out of {arr_s.shape[0]}. "
                        f"Your ToA array has therefore been filled within {round(100 * i_sample/arr_s.shape[0])} % "
                        "of the signal."
                    )
                # Interpolate the ToA
                arr_toa[i_toa] = arr_t[i_sample - 1] + (
                    arr_t[i_sample] - arr_t[i_sample - 1]
                ) * (threshold - prev_sample) / (curr_sample - prev_sample)
                i_toa += 1

        # Update the previous sample value
        prev_sample = curr_sample

    # Return the array containing the ToAs
    return arr_toa[:i_toa]


@njit
def threshold_crossing_hysteresis_rising(
    arr_t: np.ndarray,
    arr_s: np.ndarray,
    threshold: float,
    hysteresis_height: float,
    n_est: Optional[float] = None,
) -> np.ndarray:
    """A sequential threshold crossing algorithm that interpolates
        the ToA between the two samples where the signal crosses
        the threshold.

    Args:
        arr_t (np.ndarray): The array containing the time values.
        arr_s (np.ndarray): The array containing the signal voltage values
            corresponding to the time values.
        threshold (float): The threshold value.
        hysteresis_height (float): The height of the hysteresis, in the same
            units as the signal.
        n_est (float, optional): The estimated number of ToAs in this
            signal. Defaults to None. This number is used to pre-allocate the array
            containing the ToAs. If this number is not provided, the array will
            be pre-allocated as the same dimension as arr_t and arr_s.

    Returns:
        np.ndarray: An array containing the ToAs.
    """
    threshold_lower = threshold - hysteresis_height
    trigger_state = True if arr_s[0] > threshold_lower else False

    # Pre-allocate the array containing the ToAs
    if n_est is None:
        arr_toa = -1 * np.ones(arr_t.shape)
    else:
        arr_toa = -1 * np.ones(n_est)

    # Initialise the index of the ToA array
    i_toa = 0

    # Initialise the previous sample value
    prev_sample = arr_s[0]

    # Loop through all the samples
    for i_sample in range(1, arr_s.shape[0]):
        # Get the current sample value
        curr_sample = arr_s[i_sample]

        # Check if the threshold is crossed
        if trigger_state is True:
            if curr_sample <= threshold_lower:
                trigger_state = False
        else:
            if curr_sample >= threshold:
                trigger_state = True
                # Interpolate the ToA
                arr_toa[i_toa] = arr_t[i_sample - 1] + (
                    arr_t[i_sample] - arr_t[i_sample - 1]
                ) * (threshold - prev_sample) / (curr_sample - prev_sample)
                i_toa += 1

        # Update the previous sample value
        prev_sample = curr_sample

    # Return the array containing the ToAs
    return arr_toa[:i_toa]

# Untested
@njit
def threshold_crossing_hysteresis_falling(
    arr_t : np.ndarray,
    arr_s : np.ndarray,
    threshold : float,
    hysteresis_height : float,
    n_est : Optional[float] = None,
) -> np.ndarray:
    """ This function implements the constant threshold triggering
    method with hysteresis on the falling edge. The hysteresis
    height is specified in the same units as the signal.

    Args:
        arr_t (np.ndarray): The time values of the signal.
        arr_s (np.ndarray): The signal to determine the threshold
            level for.
        threshold (float): The threshold level to use for the
            constant threshold triggering method.
        hysteresis_height (float): The height of the hysteresis.
            It has the same units as the signal.
        n_est (Optional[float]): The estimated number of ToAs in
            this signal. Defaults to None. This number is used to
            pre-allocate the array containing the ToAs. If this
            number is not provided, the array will be pre-allocated
            as the same dimension as arr_t and arr_s. You should
            specify this value for large signals.
    """
    threshold_upper = threshold + hysteresis_height
    trigger_state = True if arr_s[0] < threshold_upper else False

    # Pre-allocate the array containing the ToAs
    if n_est is None:
        arr_toa = -1 * np.ones(arr_t.shape)
    else:
        arr_toa = -1 * np.ones(n_est)

    # Initialise the index of the ToA array
    i_toa = 0

    # Initialise the previous sample value
    prev_sample = arr_s[0]

    # Loop through all the samples
    for i_sample in range(1, arr_s.shape[0]):
        # Get the current sample value
        curr_sample = arr_s[i_sample]

        # Check if the threshold is crossed
        if trigger_state is True:
            if curr_sample >= threshold_upper:
                trigger_state = False
        else:
            if curr_sample <= threshold:
                trigger_state = True
                # Interpolate the ToA
                arr_toa[i_toa] = (
                    arr_t[i_sample - 1] 
                    + (arr_t[i_sample] - arr_t[i_sample - 1]) 
                    * (threshold - prev_sample) 
                    / (curr_sample - prev_sample)
                )
                i_toa += 1

        # Update the previous sample value
        prev_sample = curr_sample

    # Return the array containing the ToAs
    return arr_toa[:i_toa]


# Copyright (c) ARAMI Physical Asset Management Pty Ltd T/A Bladesight Inc. (2023)

from typing import List, Tuple, Union, Literal

import numpy as np
from numba import njit
import duckdb

import pandas as pd
import polars as pl

from ..utils import _get_dataframe_library_preference


def transform_ToAs_to_AoAs(
    df_opr_zero_crossings: Union[np.ndarray, pd.DataFrame, pl.DataFrame],
    df_probe_toas: Union[np.ndarray, pd.DataFrame, pl.DataFrame],
) -> Union[pd.DataFrame, pl.DataFrame]:
    """This function transforms the ToA values to AoA values for a
    single probe, given the OPR zero-crossing times and the proximity
    probe's ToA values. It receives Pandas DataFrames, and also
    cleans up the ToAs that could not be converted.

    The timestamps are assumed to reside in the first column of
    each DataFrame.

    Args:
        df_opr_zero_crossings (Union[np.ndarray, pd.DataFrame, pl.DataFrame]): The OPR zero-crossing
            times. If a DataFrame is passed, the first column is assumed to contain the timestamps.
        df_probe_toas (Union[np.ndarray, pd.DataFrame, pl.DataFrame]): The proximity probe's ToA values.
            If a DataFrame is passed, the first column is assumed to contain the timestamps.

    Returns:
        Union[pd.DataFrame, pl.DataFrame]: A DataFrame with the AoA values.
    """
    if not isinstance(df_opr_zero_crossings, (np.ndarray, pd.DataFrame, pl.DataFrame)):
        raise ValueError("df_opr_zero_crossings must be a numpy array or a DataFrame.")
    if not isinstance(df_probe_toas, (np.ndarray, pd.DataFrame, pl.DataFrame)):
        raise ValueError("df_probe_toas must be a numpy array or a DataFrame.")
    if isinstance(df_opr_zero_crossings, np.ndarray):
        df_opr_zero_crossings = pl.DataFrame({'toas':df_opr_zero_crossings})
    if isinstance(df_probe_toas, np.ndarray):
        df_probe_toas = pl.DataFrame({'toas':df_probe_toas})
    first_column_opr = duckdb.execute(
        "SELECT * FROM df_opr_zero_crossings LIMIT 0"
    ).pl().columns[0]
    
    first_column_probe = duckdb.execute(
        "SELECT * FROM df_probe_toas LIMIT 0"
    ).pl().columns[0]
    
    query = f"""
    WITH TAB1 AS (
        SELECT 
            (row_number() OVER () - 1)::UINTEGER  as n,
            {first_column_opr} AS n_start_time,
            LEAD({first_column_opr}) OVER (ORDER BY {first_column_opr}) AS n_end_time,
            2 * PI()/(n_end_time - n_start_time) AS Omega
        FROM 
            df_opr_zero_crossings
    )

    SELECT 
        TAB1.*,
        df_probe_toas.{first_column_probe} AS ToA,
        (
            df_probe_toas.{first_column_probe} 
            - TAB1.n_start_time
        ) * Omega AS AoA,
    FROM df_probe_toas
    LEFT JOIN TAB1 
    ON df_probe_toas.{first_column_probe} >= TAB1.n_start_time
    WHERE 
        Omega IS NOT NULL
    AND
        ToA <= n_end_time
    """
    if _get_dataframe_library_preference() == 'pd':
        return duckdb.execute(query).df()
    elif _get_dataframe_library_preference() == 'pl':
        return duckdb.execute(query).pl()
    else:
        raise ValueError(
            "The DataFrame library preference must be either 'pd' or 'pl'."
            " Please set it using the bladesight.utils.set_dataframe_library_preference function."
        )

def transform_ToAs_to_AoAs_mpr(
    df_ias: Union[pd.DataFrame, pl.DataFrame],
    df_probe_toas: Union[pd.DataFrame, pl.DataFrame]
) -> Union[pd.DataFrame, pl.DataFrame]:
    """This function transforms the ToA values to AoA values for a
    single probe, given the IAS DataFrame.

    The timestamps are assumed to reside in the first column of
    df_probe.

    Args:
        df_ias (Union[pd.DataFrame, pl.DataFrame]): The result of the 
            bladesight.ias.calculate_ias function.
        df_probe_toas (Union[pd.DataFrame, pl.DataFrame]): A DataFrame with the probe's
            ToAs.
    Returns:
        pd.DataFrame: A DataFrame containing the AoA values.
    """
    if not isinstance(df_ias, (pd.DataFrame, pl.DataFrame)):
        raise TypeError(
            "df_ias must be a Pandas or Polars DataFrame."
        )
    if not isinstance(df_probe_toas, (pd.DataFrame, pl.DataFrame)):
        raise TypeError(
            "df_probe_toas must be a Pandas or Polars DataFrame."
        )
    present_columns = duckdb.execute(
        "SELECT * FROM df_ias LIMIT 0"
    ).pl().columns

    if present_columns != [
        'section_start_time',
        'section_end_time',
        'section_distance',
        'Omega',
        'n',
        'section_start',
        'section_end'
    ]:
        raise AssertionError(
            """The DataFrame must have the columns: 'section_start_time', """
            """ 'section_end_time', 'section_distance', 'Omega', 'n', """ 
            """ 'section_start', 'section_end'"""
        )
    left_merge_key = duckdb.execute(
        "SELECT * FROM df_probe_toas LIMIT 0"
    ).pl().columns[0]
    query = f"""
        SELECT 
            ias.n,
            ias.section_start_time,
            ias.section_end_time,
            ias.Omega,
            probe.toa as ToA,
            ((probe.toa - ias.section_start_time) * ias.Omega + ias.section_start) % (2*PI()) as AoA
        FROM df_probe_toas probe
        ASOF LEFT JOIN 
            df_ias ias
            ON {left_merge_key} >= section_start_time
        """
    if _get_dataframe_library_preference() == 'pl':
        return duckdb.execute(query).pl()
    elif _get_dataframe_library_preference() == 'pd':
        return duckdb.execute(query).df()
    else:
        raise ValueError(
            "The DataFrame library preference must be either 'pd' or 'pl'."
            " Please set it using the bladesight.utils.set_dataframe_library_preference function."
        )

##########################################################################
#                    TRANSFORM PROX AoAs to Blade AoAs                   #
##########################################################################


def calculate_Q(
    arr_aoas: Union[np.ndarray, pd.Series, pl.Series], 
    d_theta: float,
    N: int
) -> Tuple[float, np.ndarray]:
    """This function calculates the binning quality factor, Q, 
    given the AoA values, the number og blades and the bin offset.

    Args:
        arr_aoas (np.ndarray): The proximity probe AoA values.
            If a Pandas Series or Polars Series is passed, the values
            are extracted.
        d_theta (float): The bin offset.
        N (int): The number of blades on the rotor.

    Returns:
        Tuple[float, np.ndarray]: The binning quality factor, Q, and
        the bin edges.
    """
    if isinstance(arr_aoas, (pd.Series, pl.Series)):
        arr_aoas = arr_aoas.to_numpy()
    bin_edges = np.linspace(0 + d_theta, 2 * np.pi + d_theta, N + 1)
    Q = 0
    for b in range(N):
        left_edge = bin_edges[b]
        right_edge = bin_edges[b + 1]
        bin_centre = (left_edge + right_edge) / 2
        bin_mask = (arr_aoas > left_edge) & (arr_aoas <= right_edge)
        Q += np.sum((arr_aoas[bin_mask] - bin_centre) ** 2)
    if np.sum(arr_aoas < bin_edges[0]) > 0:
        # In other words, some values are less than bin_edges[0],
        # meaning it actually occurs at the end of the previous
        # revolution
        left_edge_last = bin_edges[N - 1]
        right_edge_last = bin_edges[N]
        bin_centre_last = (left_edge_last + right_edge_last) / 2
        bin_mask = arr_aoas <= bin_edges[0]
        Q += np.sum(((2 * np.pi - arr_aoas[bin_mask]) - bin_centre_last) ** 2)
    if np.sum(arr_aoas > bin_edges[-1]) > 0:
        left_edge_first = bin_edges[0]
        right_edge_first = bin_edges[1]
        bin_centre_first = (left_edge_first + right_edge_first) / 2
        bin_mask = arr_aoas > bin_edges[-1]
        Q += np.sum(((arr_aoas[bin_mask] - 2 * np.pi) - bin_centre_first) ** 2)
    return Q, bin_edges

def transform_prox_AoAs_to_blade_AoAs(
    df_prox: Union[pd.DataFrame, pl.DataFrame],
    B: int,
    d_theta_increments: int = 200,
    spurious_pulse_policy : Literal['drop', 'keep'] = 'drop'
) -> List[Union[pd.DataFrame, pl.DataFrame]]:
    """This function takes a DataFrame containing the AoA values of a proximity probe,
    and returns a list of DataFrame, each containing the AoA values of a single blade.

    Args:
        df_prox (Union[pd.DataFrame, pl.DataFrame]): The dataframe containing the AoA values
            of the proximity probe.
        B (int): The number of blades.
        d_theta_increments (int, optional): The number of increments.
        spurious_pulse_policy (Literal['drop', 'keep'], optional): The policy to handle
            spurious pulses. Defaults to 'drop'.

    Returns:
        List[Union[pd.DataFrame, pl.DataFrame]]: A list of dataframes, each containing the AoA
        values of a single blade.
    """
    return_type = _get_dataframe_library_preference()

    d_thetas = np.linspace(-0.5 * np.pi / B, 1.5 * np.pi / B, d_theta_increments)
    arr_aoas = duckdb.execute("SELECT AoA FROM df_prox").fetchnumpy()["AoA"]    
    Qs = []
    optimal_Q, optimal_bin_edges, optimal_d_theta = np.inf, None, None
    for d_theta in d_thetas:
        Q, bin_edges = calculate_Q(arr_aoas, d_theta, B)
        if Q < optimal_Q:
            optimal_Q = Q * 1
            optimal_bin_edges = bin_edges
            optimal_d_theta = d_theta * 1
        Qs.append(Q)

    blade_dfs = []
    blade_median_AoAs = []
    for b in range(B):
        if b == 0:
            query = f"""
                SELECT 
                    * 
                FROM df_prox
                WHERE
                    (
                        (AoA > {optimal_bin_edges[b]})
                            AND
                        (AoA <= {optimal_bin_edges[b + 1]})
                    )
                OR (
                    AoA > {optimal_bin_edges[-1]}
                )
                ORDER BY ToA
            """
        elif b == B - 1:
            query = f"""
                WITH TAB1 AS (
                    SELECT 
                        * 
                    FROM df_prox
                    WHERE
                        (
                            (AoA > {optimal_bin_edges[b]})
                                AND
                            (AoA <= {optimal_bin_edges[b + 1]})
                        )
                    OR (
                        AoA <= {optimal_bin_edges[0]}
                    )
                    ORDER BY ToA
                )
                SELECT
                    * EXCLUDE (AoA),
                    CASE
                        WHEN AoA > {optimal_bin_edges[-1]} THEN 2 * PI() - AoA
                        ELSE AoA
                    END AS AoA
                FROM TAB1
            """
        else:
            query = f"""
                SELECT 
                    * 
                FROM df_prox
                WHERE
                    (
                        (AoA > {optimal_bin_edges[b]})
                        AND
                        (AoA <= {optimal_bin_edges[b + 1]})
                    )
                ORDER BY ToA
            """
        if return_type == 'pd':
            df_bin = duckdb.execute(query).df()
        else:
            df_bin = duckdb.execute(query).pl()
        blade_dfs.append(df_bin)
        blade_median_AoAs.append(
            duckdb.execute("SELECT AoA.MEDIAN() as median FROM df_bin").fetchone()[0]
        )

    blade_order = np.argsort(blade_median_AoAs)
    blade_dfs = [blade_dfs[i] for i in blade_order]
    
    if spurious_pulse_policy == 'keep':
        return blade_dfs
    
    # Drop revolutions that has more than one occurrence of a blade
    blade_dfs_to_return = []
    query = """
        SELECT * FROM df
        QUALIFY row_number() OVER (PARTITION BY n ORDER BY ToA) = 1
        ORDER BY n, ToA
    """
    for _, df in enumerate(blade_dfs):
        if return_type == 'pd':
            blade_dfs_to_return.append(duckdb.execute(query).df())
        else:
            blade_dfs_to_return.append(duckdb.execute(query).pl())
    return blade_dfs_to_return



from typing import List, Union
import pandas as pd
import polars as pl

from .aoa import (
    transform_ToAs_to_AoAs, 
    transform_ToAs_to_AoAs_mpr, 
    transform_prox_AoAs_to_blade_AoAs
)
from .align import pivot_blade_AoAs_along_revolutions, assemble_rotor_AoA_dfs
from .zero import get_blade_tip_deflections_from_AoAs

__all__ = [
    "get_rotor_blade_AoAs",
    "get_blade_tip_deflections_from_AoAs"
]

def get_rotor_blade_AoAs(
    df_encoder : pd.DataFrame,
    prox_probe_toas : List[Union[pd.DataFrame, pl.DataFrame]],
    probe_spacings : List[float],
    B : int,
    is_mpr : bool = False
) -> List[pd.DataFrame]:
    """This function converts the raw time stamps, both the OPR zero-crossing
    times and he proximity probe ToAs, and returns a DataFrame for each 
    blade containing the ToA and AoA values for each blade over all the
    proximity probes.

    Args:
        df_encoder (pd.DataFrame): A DataFrame containing the
            zero-crossing times and corresponding info. This DataFrame can
            take on two values: 
                1) OPR encoder: The first column of the DataFrame should be
                the zero-crossing times of the OPR encoder.
                2) MPR encoder: The DataFrame MUST be the result of the
                bladesight.ias.calculate_mpr function.
        prox_probe_toas (List[pd.DataFrame]): A list of DataFrames
            where each DataFrame contains the ToAs of a single
            blade from a proximity probe.
        probe_spacings (List[float]): The relative spacing between
            the first probe and every other probe. There are one
            less value in this list than in prox_probe_toas.
        B (int): The number of blades.
        is_mpr (bool, optional): A flag to indicate if the encoder is
            an MPR encoder. Defaults to False.

    Returns:
        List[pd.DataFrame]: A list of DataFrames where each DataFrame
            contains the ToAs and AoAs of a single blade over all
            the proximity probes.
    """
    blade_dfs_recombined = []

    if not is_mpr:
        for df_prox_toas in prox_probe_toas:
            df_prox = transform_ToAs_to_AoAs(
                df_encoder.to_frame() if isinstance(df_encoder, (pd.Series, pl.Series)) else df_encoder, 
                df_prox_toas.to_frame() if isinstance(df_prox_toas, (pd.Series, pl.Series)) else df_prox_toas
            )
            
            blade_dfs_recombined.append(
                pivot_blade_AoAs_along_revolutions(
                    transform_prox_AoAs_to_blade_AoAs(
                        df_prox, 
                        B
                    )
                )
            )
    else:
        for df_prox_toas in prox_probe_toas:
            try:
                df_prox = transform_ToAs_to_AoAs_mpr(
                    df_encoder.to_frame() if isinstance(df_encoder, (pd.Series, pl.Series)) else df_encoder, 
                    df_prox_toas.to_frame() if isinstance(df_prox_toas, (pd.Series, pl.Series)) else df_prox_toas
                )
            except AssertionError as e:
                print("It looks like you did not pass the correct MPR encoder DataFrame.")
                print("Please ensure you pass df_encoder as the result of the bladesight.ias.calculate_mpr function.")
                raise e

            blade_dfs_recombined.append(
                pivot_blade_AoAs_along_revolutions(
                    transform_prox_AoAs_to_blade_AoAs(
                        df_prox, 
                        B
                    )
                )
            )
    
    rotor_AoA_dfs = assemble_rotor_AoA_dfs(
        prox_aligned_dfs=blade_dfs_recombined,
        probe_spacing=probe_spacings
    )
    return rotor_AoA_dfs


from bladesight.dataset_handler import (
    get_path_to_local_bladesight,
    get_local_datasets,
    get_bladesight_datasets,
    _confirm_dataset_is_valid,
    _execute_sql_without_arg,
    _get_db_tables,
    _get_all_metadata,
    _get_printable_citation,
    Dataset,
    BladesightDatasetDirectory
)
from bladesight.utils import (
    _get_dataframe_library_preference, 
    set_dataframe_library_preference
)
from pathlib import Path
import os
import pytest
import duckdb
from duckdb import CatalogException
import pandas as pd
import polars as pl
from polars.testing import assert_frame_equal
import json

def testget_path_to_local_bladesight():
    bladesight_data_root_path = str(get_path_to_local_bladesight())
    assert str(bladesight_data_root_path).endswith(".bladesight")
    os.environ["BLADESIGHT_DATASETS_PATH"] = "/user/defined/path/"
    bladesight_data_root_path = get_path_to_local_bladesight()
    assert bladesight_data_root_path == Path("/user/defined/path/.bladesight")

def test_get_local_datasets(tmp_path):
    os.environ["BLADESIGHT_DATASETS_PATH"] = str(tmp_path)
    assert get_local_datasets() == []
    path_to_bladesight = get_path_to_local_bladesight()
    path_to_bladesight.mkdir(parents=True, exist_ok=True)
    assert get_local_datasets() == []
    path_to_folder1 = path_to_bladesight / "folder1"
    assert get_local_datasets() == []
    path_to_folder1.mkdir(parents=True, exist_ok=True)
    assert get_local_datasets() == []
    # Create a .txt file in get_path_to_local_bladesight()
    with open(path_to_folder1 / "test.txt", 'w') as f:
        f.write("test")
    assert get_local_datasets() == []
    # Now write a file with a .db extention
    with open(path_to_folder1 / "test.db", "w") as f:
        f.write("test")
    assert get_local_datasets() == ["folder1/test"]

def test_get_bladesight_datasets():
    """ This is terrible, I know
    """
    minimum_sets_present = [
        'bladesight-datasets/intro_to_btt/intro_to_btt_ch02', 
        'bladesight-datasets/intro_to_btt/intro_to_btt_ch03', 
        'bladesight-datasets/intro_to_btt/intro_to_btt_ch05', 
        'bladesight-datasets/intro_to_btt/intro_to_btt_ch06'
    ]
    bladesight_online_sets = get_bladesight_datasets()
    assert all([dataset in bladesight_online_sets for dataset in minimum_sets_present])

def test_confirm_dataset_is_valid(tmp_path):
    tmp_path.mkdir(parents=True, exist_ok=True)
    path_to_file = tmp_path / "test.txt"
    # Check that we get a FileNotFoundError if the file doesn't exist
    with pytest.raises(FileNotFoundError):
        _confirm_dataset_is_valid(path_to_file)
    # Create a file, now we should get a ValueError because 
    # the extention
    # is not .db
    path_to_file.touch()
    with pytest.raises(ValueError):
        _confirm_dataset_is_valid(path_to_file)
    # Now write a file with a .db extention
    path_to_file = tmp_path / "test.db"
    path_to_file.touch()
    assert _confirm_dataset_is_valid(path_to_file) is None

def test___execute_sql_without_arg(tmp_path):
    tmp_path.mkdir(parents=True, exist_ok=True)
    path_to_file = tmp_path / "test.db"
    with duckdb.connect(str(path_to_file)) as _:
        pass
    df_empty_tables_pd = _execute_sql_without_arg(path_to_file, "SHOW TABLES;")
    assert df_empty_tables_pd.shape == (0, 1)
    assert df_empty_tables_pd.columns[0] == "name"
    assert isinstance(df_empty_tables_pd, pd.DataFrame)
    df_empty_tables_pl = _execute_sql_without_arg(path_to_file, "SHOW TABLES;", "pl")
    assert df_empty_tables_pl.shape == (0, 1)
    assert df_empty_tables_pl.columns[0] == "name"
    assert isinstance(df_empty_tables_pl, pl.DataFrame)
    with pytest.raises(ValueError):
        _execute_sql_without_arg(path_to_file, "SHOW TABLES;", "invalid")
    df_table_1 = pl.DataFrame({
        "A" : [1,2,3,4],
        "B" : [5,6,7,8]
    })
    df_table_2 = pl.DataFrame({
        "A" : [9,10,11,12],
        "B" : [13,14,15,16]
    })
    with duckdb.connect(str(path_to_file)) as con:
        con.execute("CREATE TABLE table_1 AS SELECT * FROM df_table_1")
        con.execute("CREATE TABLE table_2 AS SELECT * FROM df_table_2")
    df_tables = _execute_sql_without_arg(path_to_file, "SHOW TABLES;")
    assert df_tables.shape == (2, 1)
    assert df_tables.columns[0] == "name"
    assert df_tables["name"].to_list() == ["table_1", "table_2"]
    assert_frame_equal(
        _execute_sql_without_arg(path_to_file, "SELECT * FROM table_1;","pl"), 
        df_table_1
    )
    assert_frame_equal(
        _execute_sql_without_arg(path_to_file, "SELECT * FROM table_2;","pl"), 
        df_table_2
    )

def test__get_db_tables(tmp_path):
    tmp_path.mkdir(parents=True, exist_ok=True)
    path_to_file = tmp_path / "test.db"
    with duckdb.connect(str(path_to_file)) as _:
        pass
    assert _get_db_tables(path_to_file) == []
    df_table_1 = pl.DataFrame({
        "A" : [1,2,3,4],
        "B" : [5,6,7,8]
    })
    df_table_2 = pl.DataFrame({
        "A" : [9,10,11,12],
        "B" : [13,14,15,16]
    })
    with duckdb.connect(str(path_to_file)) as con:
        con.execute("CREATE TABLE table_1 AS SELECT * FROM df_table_1")
        con.execute("CREATE TABLE table_2 AS SELECT * FROM df_table_2")
    assert _get_db_tables(path_to_file) == ["table_1", "table_2"]
    df_metadata = pl.DataFrame({
        "A" : [1,2,3,4],
        "B" : [5,6,7,8]
    })
    with duckdb.connect(str(path_to_file)) as con:
        con.execute("CREATE TABLE metadata AS SELECT * FROM df_metadata")
    assert _get_db_tables(path_to_file) == ["table_1", "table_2"]

def test__get_all_metadata(tmp_path):
    tmp_path.mkdir(parents=True, exist_ok=True)
    path_to_file = tmp_path / "test.db"
    with duckdb.connect(str(path_to_file)) as _:
        pass
    with pytest.raises(CatalogException):
        _get_all_metadata(path_to_file)
    json_object_1 = {
        "A" : [1,2,3,4],
        "B" : [5,6,7,8]
    }
    json_object_2 = {
        "CITATION" : """@article{diamond2021constant,
            title={Constant speed tip deflection determination using the instantaneous phase of blade tip timing data},
            author={Diamond, DH and Heyns, Philippus Stephanus and Oberholster, AJ},
            journal={Mechanical Systems and Signal Processing},
            volume={150},
            pages={107151},
            year={2021},
            publisher={Elsevier}
            }""",
        "URL" : """https://repository.up.ac.za/bitstream/handle/2263/86905/Diamond_Constant_2021.pdf?sequence=1""" 
    }
    
    df_metadata = pl.DataFrame({
        "metadata_key" : ["A", "B"],
        "metadata_value" : [json.dumps(json_object_1), json.dumps(json_object_2)]
    })
    with duckdb.connect(str(path_to_file)) as con:
        con.execute("CREATE TABLE metadata AS SELECT * FROM df_metadata")
    metadata = _get_all_metadata(path_to_file)
    assert metadata["A"] == json_object_1
    assert metadata["B"] == json_object_2

def test__get_printable_citation():
    metadata_dict = {}
    assert _get_printable_citation(metadata_dict) == "No citation provided in metadata table."
    metadata_dict = {
        "CITATIONs" : {}
    }
    assert _get_printable_citation(metadata_dict) == "No citation provided in metadata table."
    metadata_dict = {
        "CITATION" : "test"
    }
    assert _get_printable_citation(metadata_dict) == "Citation format error. Here is the raw citation:\n\n" + str(metadata_dict["CITATION"])
    metadata_dict = {
        "CITATION" : {
            "title" : "test"
        }
    }
    assert _get_printable_citation(metadata_dict) == "Citation format error. Here is the raw citation:\n\n" + str(metadata_dict["CITATION"])
    metadata_dict = {
        "CITATION" : {
            "repr" : "uber test"
        }
    }
    assert _get_printable_citation(metadata_dict) == f"""If you use this dataset in published work, please use the below citation:\n\n{metadata_dict['CITATION']['repr']}"""
    metadata_dict = {
        "CITATION" : {
            "repr" : "uber test",
            "url" : "https://test.com",            
        }
    }
    assert_test = f"""If you use this dataset in published work, please use the below citation:\n\n{metadata_dict['CITATION']['repr']}""" + f"""\nLink to paper: {metadata_dict['CITATION']['url']}"""
    assert _get_printable_citation(metadata_dict) == assert_test
    metadata_dict = {
        "CITATION" : {
            "repr" : "uber test",
            "url" : "https://test.com",
            "doi" : "10.1234/5678"
        }
    }
    assert_test += f"""\nDOI: {metadata_dict['CITATION']['doi']}"""
    assert _get_printable_citation(metadata_dict) == assert_test

def test_Dataset(tmp_path):
    tmp_path.mkdir(parents=True, exist_ok=True)
    path_to_file = tmp_path / "test.db"
    with duckdb.connect(str(path_to_file)) as _:
        pass
    df_metadata = pl.DataFrame({
        "metadata_key" : ["CITATION"],
        "metadata_value" : [
            json.dumps({
                "repr" : "uber test",
                "url" : "https://test.com",
                "doi" : "10.1234/5678"
            })
        ]
    })
    df_test_1 = pl.DataFrame({
        "A" : [1,2,3,4],
        "B" : [5,6,7,8]
    })
    df_test_2 = pl.DataFrame({
        "A" : [9,10,11,12],
        "B" : [13,14,15,16]
    })
    with duckdb.connect(str(path_to_file)) as con:
        con.execute("CREATE TABLE metadata AS SELECT * FROM df_metadata")
        con.execute("CREATE TABLE test_1 AS SELECT * FROM df_test_1")
        con.execute("CREATE TABLE test_2 AS SELECT * FROM df_test_2")

    ds = Dataset(path_to_file)
    assert _get_dataframe_library_preference() == 'pd'
    set_dataframe_library_preference('pl')
    assert _get_dataframe_library_preference() == 'pl'
    set_dataframe_library_preference('pd')
    assert _get_dataframe_library_preference() == 'pd'
    with pytest.raises(ValueError):
        set_dataframe_library_preference('invalid')
    assert sorted(ds._ipython_key_completions_()) == ["table/test_1", "table/test_2"]
    repr = ds.__repr__()
    assert "table/test_1" in repr
    assert "table/test_2" in repr
    assert sorted(ds.tables) == ["test_1", "test_2"]
    
    set_dataframe_library_preference('pl')
    df_test_1_read = ds["table/test_1"]
    assert_frame_equal(df_test_1, df_test_1_read)
    df_test_2_read = ds["table/test_2"]
    assert_frame_equal(df_test_2, df_test_2_read)
    assert ds.metadata["CITATION"]["repr"] == "uber test"
    assert ds.metadata["CITATION"]["url"] == "https://test.com"
    assert ds.metadata["CITATION"]["doi"] == "10.1234/5678"

def test_replace_path_prefix_static_method():
    old_path = "bladesight-datasets/intro_to_btt/intro_to_btt_ch02"

    new_path = BladesightDatasetDirectory._replace_path_prefix(
        old_path,
        "data"
    )
    assert new_path == "data/intro_to_btt/intro_to_btt_ch02"

def test__getitem_key_correct_format():
    assert BladesightDatasetDirectory._getitem_key_correct_format("data/test_1")
    assert BladesightDatasetDirectory._getitem_key_correct_format("table/test_1/") is False


import numpy as np
import pandas as pd
from typing import List, Dict, Union

def cff_method_single_revolution(
    df_blade : pd.DataFrame,
    theta_sensor_set : List[float],
    EO : int,
    signal_suffix : str = "_filt" 
) -> pd.DataFrame:
    """ This function fits the CFF method for a resonance by 
    using a single revolution of data for each set of CFF parameters.

    Args:
        df_blade (pd.DataFrame): The dataframe containing the tip deflections.
        theta_sensor_set (List[float]): The sensor angles.
        EO (int): The Engine Order.
        signal_suffix (str, optional): The suffix of the tip deflection 
            signals. Defaults to "_filt".

    Returns:
        pd.DataFrame: A DataFrame containing the CFF parameters for 
            each shaft revolution. 
    """
    PROBE_COUNT = len(theta_sensor_set)
    tip_deflection_signals = [
        f"x_p{i_probe + 1}{signal_suffix}" 
        for i_probe in range(PROBE_COUNT)
    ]
    theta_sensors = np.array(theta_sensor_set)
    A = np.ones((PROBE_COUNT, 3))
    A[:, 0] = np.sin(theta_sensors * EO)
    A[:, 1] = np.cos(theta_sensors * EO)

    A_pinv = np.linalg.pinv(A) 
    B = A_pinv.dot(
        df_blade.loc[:, tip_deflection_signals].values.T
    ) 
    df_cff = pd.DataFrame(B.T, columns=["A", "B", "C"]) 
    df_cff["X"] = np.sqrt(df_cff["A"]**2 + df_cff["B"]**2)
    df_cff["phi"] = np.arctan2(df_cff["A"], df_cff["B"])
    df_cff["n"] = df_blade["n"].values
    df_predicted_targets = pd.DataFrame(
        A.dot(B).T, 
        columns=[
            col + "_pred" 
            for col 
            in tip_deflection_signals
        ]
    ) 
    df_cff = pd.concat([df_cff, df_predicted_targets], axis=1)
    return df_cff

def cff_method_multiple_revolutions(
    df_blade : pd.DataFrame,
    theta_sensor_set : List[float],
    EO : int,
    extra_revolutions : int,
    signal_suffix : str = "_filt" 
) -> pd.DataFrame:
    """ This function fits the CFF method for a resonance by
    using multiple revolutions of data for each set of CFF parameters.

    Args:
        df_blade (pd.DataFrame): The dataframe containing the tip deflections.
        theta_sensor_set (List[float]): The sensor angles.
        EO (int): The Engine Order.
        extra_revolutions (int): The number of revolutions to use for the fit.
        signal_suffix (str, optional): The suffix of the tip deflection 
            signals. Defaults to "_filt".

    Returns:
        pd.DataFrame: A DataFrame containing the CFF parameters for 
            each shaft revolution.
    """
    PROBE_COUNT = len(theta_sensor_set)
    tip_deflection_signals = [
        f"x_p{i_probe + 1}{signal_suffix}" 
        for i_probe in range(PROBE_COUNT)
    ]
    theta_sensors = np.array(theta_sensor_set)

    A = np.ones((PROBE_COUNT*(2*extra_revolutions+1), 3))
    arr_multiple_thetas = np.array(
        list(theta_sensors)*(2*extra_revolutions+1)
    )
    A[:, 0] = np.sin(arr_multiple_thetas * EO)
    A[:, 1] = np.cos(arr_multiple_thetas * EO)
    A_pinv = np.linalg.pinv(A)
    new_obs_rows = df_blade.shape[0] - 2*extra_revolutions
    X_multiple_revos = np.zeros(
        (
            new_obs_rows, 
            PROBE_COUNT*(2*extra_revolutions+1)
        )
    )
    for n_revo in range(-extra_revolutions, extra_revolutions+1):
        for i_probe in range(PROBE_COUNT):
            mat_aoas_start = extra_revolutions + n_revo
            mat_aoas_end = mat_aoas_start + new_obs_rows
            i_col = i_probe + n_revo*PROBE_COUNT + extra_revolutions*PROBE_COUNT
            X_multiple_revos[:,i_col] = (
                df_blade.iloc[mat_aoas_start:mat_aoas_end][tip_deflection_signals[i_probe]]
            )
    B = A_pinv.dot(X_multiple_revos.T)
    B_full = np.zeros((df_blade.shape[0], 3))
    B_full[extra_revolutions:-extra_revolutions, :] = B.T
    B_full[:extra_revolutions, :] = B_full[extra_revolutions, :]
    B_full[-extra_revolutions:, :] = B_full[-extra_revolutions-1, :]

    df_cff = pd.DataFrame(B_full, columns=["A", "B", "C"])
    df_cff["X"] = np.sqrt(df_cff["A"]**2 + df_cff["B"]**2)
    df_cff["phi"] = np.arctan2(df_cff["A"], df_cff["B"])
    df_cff["n"] = df_blade["n"].values
    target_matrix = (A.dot(B_full.T)).T
    predicted_deflections = target_matrix[:, extra_revolutions*PROBE_COUNT:(extra_revolutions+1)*PROBE_COUNT] 
    df_predicted_targets = pd.DataFrame(
        predicted_deflections, 
        columns=[
            col + "_pred" 
            for col 
            in tip_deflection_signals
        ]
    )
    df_cff = pd.concat([df_cff, df_predicted_targets], axis=1)
    return df_cff

def perform_CFF_fit(
    df_blade : pd.DataFrame,
    n_start : int,
    n_end : int,
    EOs : List[int] = np.arange(1, 20),
    extra_revolutions : int = 1
) -> Dict[str, Union[pd.DataFrame, int]]:
    """ This function performs the CFF method fit for a resonance. The function
    iterates over EOs and selects the EO that gives the lowest sum of squared
    errors between the measured and predicted tip deflections.

    Args:
        df_blade (pd.DataFrame): The dataframe containing the tip deflections.
        n_start (int): The start revolution number.
        n_end (int): The end revolution number.
        EOs (List[int], optional): The EOs to consider for this resonance. Defaults 
            to np.arange(1, 20).
        extra_revolutions (int, optional): How many extra revolutions to use for 
            the fit. Defaults to 1.

    Returns:
        Dict[str, Union[pd.DataFrame, int]]: A dictionary containing the CFF 
            parameters and the selected EO.
    """
    PROBE_COUNT = len(
        [
            col 
            for col in df_blade.columns
            if col.endswith("_filt")
        ]
    )
    theta_sensor_set = [
        df_blade[f"AoA_p{i_probe + 1}"].median()
        for i_probe in range(PROBE_COUNT)
    ]
    errors = []
    df_resonance_window = df_blade.query(f"n >= {n_start} and n <= {n_end}")
    for EO in EOs:
        df_cff_params = cff_method_multiple_revolutions(
            df_resonance_window,
            theta_sensor_set,
            EO,
            extra_revolutions
        )
        error = 0
        for i_probe in range(PROBE_COUNT):
            error += np.sum(
                (
                    df_cff_params[f"x_p{i_probe+1}_filt_pred"].values 
                    - df_resonance_window[f"x_p{i_probe+1}_filt"].values
                )**2
            )
        errors.append(error)
    EO = EOs[np.argmin(errors)]
    df_cff_params = cff_method_multiple_revolutions(
        df_resonance_window,
        theta_sensor_set,
        EO,
        extra_revolutions
    )
    return {
        "df_cff_params" : df_cff_params,
        "EO" : EO
    }


from bladesight.dataset_handler import (
    get_local_datasets,
    download_dataset_from_bladesight_data,
    BladesightDatasetDirectory
)
import bladesight.dataset_handler
import os
import pytest

def test_download_dataset_from_bladesight_data(tmp_path):
    os.environ["BLADESIGHT_DATASETS_PATH"] = str(tmp_path)
    assert get_local_datasets() == []
    download_dataset_from_bladesight_data("bladesight-datasets/intro_to_btt/intro_to_btt_ch03")
    assert get_local_datasets() == ["data/intro_to_btt/intro_to_btt_ch03"]
    download_dataset_from_bladesight_data("bladesight-datasets/intro_to_btt/intro_to_btt_ch05")
    assert get_local_datasets() == ["data/intro_to_btt/intro_to_btt_ch03", "data/intro_to_btt/intro_to_btt_ch05"]

def test_Datasets(tmp_path, monkeypatch):
    """This function will test whether we can return 
    datasets that have been downloaded. 
    """
    os.environ["BLADESIGHT_DATASETS_PATH"] = str(tmp_path)
    Datasets = BladesightDatasetDirectory()
    ds = Datasets["data/intro_to_btt/intro_to_btt_ch03"]
    df = ds["table/du_toit_2017_test_1_opr_zero_crossings"]
    assert df.shape == (791, 1)
    assert df.columns == ["time"]
    def mock_download_dataset_from_bladesight_data(*args, **kwargs):
        raise ValueError("Mocked error")
    monkeypatch.setattr(
        bladesight.dataset_handler, 
        "download_dataset_from_bladesight_data", 
        mock_download_dataset_from_bladesight_data
    )
    Datasets = BladesightDatasetDirectory()
    ds = Datasets["data/intro_to_btt/intro_to_btt_ch03"]
    df = ds["table/du_toit_2017_test_1_opr_zero_crossings"]
    assert df.shape == (791, 1)
    with pytest.raises(ValueError):
        Datasets["data/intro_to_btt/intro_to_btt_ch05"]
    with pytest.raises(KeyError):
        Datasets["error/intro_to_btt/intro_to_btt_ch05"]



import numpy as np
from scipy.sparse.linalg import spsolve
from scipy.sparse import csr_matrix
import time
from typing import Optional
import polars as pl
from tqdm import tqdm
from numba import njit

def perform_bayesian_geometry_compensation(
    t: np.ndarray,
    N: int,
    M: int,
    e: np.ndarray = [],
    beta: Optional[
        float
    ] = 10.0e10,  # I actually think paper says the default value for beta is 1E10 not 10E10
    sigma: Optional[float] = 0.01,
) -> np.ndarray:
    """
    Perform geometry compensation on an incremental shaft encoder with N sections 
    measured over M revolutions.

    An algorithm that performs shaft encoder geometry compensation for 
    arbitrary shaft speeds using Bayesian Geometry Compensation.

    Please reference the paper discussing the algorithm as 
    described in the "References" section.

    Args:
    t (np.ndarray) : 1D numpy array of zeros crossing 
        times.  The first zero crossing time indicates the start of the 
        first section.  This array should therefore 
        have exactly M*N + 1 elements.
    N (int): The number of sections in the shaft encoder.
    M (int): The number of complete revolutions over which 
        the compensation must be performed.
    e (np.ndarray): An initial estimate for the 
        encoder geometry.  If left an empty array, all sections are assumed equal.
    beta (float):   Precision of the likelihood function.
    sigma (float):  Standard deviation of the prior probability.

    Returns:
        np.ndarray : An array containing the circumferential distances of all N sections.

    References:
    This function is adapted from the following paper,
    [1] D. H. Diamond, P. S. Heyns, and A. J. Oberholster, “Online Shaft Encoder Geometry 
        Compensation for Arbitrary Shaft Speed Profiles Using Bayesian Regression,” Mechanical Systems 
        and Signal Processing, vol. 81, pp. 402–418, Dec. 2016, doi: 10.1016/j.ymssp.2016.02.060.
    
    Copyright (C) 2016 Dawie Diamond, dawie.diamond@yahoo.com

    www.up.ac.za/caim
    www.caimlabs.co.za

    """
    if len(t) != M * N + 1:
        raise ValueError(
            "Input Error: The vector containing the zero-crossing times should contain exactly N*M + 1 values"
        )
    if len(e) != 0 and len(e) != N:
        raise ValueError(
            "Input Error The encoder input should either be an empty list or a list with N elements"
        )
    A = np.zeros((2 * M * N - 1, N + 2 * M * N))
    B = np.zeros((2 * M * N - 1, 1))
    #A = np.empty((2 * M * N - 1, N + 2 * M * N)) # Use empty and not zeros to save on memory usage
    #B = np.empty((2 * M * N - 1, 1)) # Use empty and not zeros to save on memory usage

    T = np.ediff1d(t)

    A[0, :N] = np.ones(N)
    B[0, 0] = 2 * np.pi
    deduct = 0
    for m in range(M):
        if m == M - 1:
            deduct = 1
        for n in range(N - deduct):
            nm = m * N + n
            A[1 + nm, n] = 3.0
            A[1 + nm, N + nm * 2] = -1.0 / 2 * T[nm] ** 2
            A[1 + nm, N + nm * 2 + 1] = -2 * T[nm]
            A[1 + nm, N + (nm + 1) * 2 + 1] = -1 * T[nm]
    deduct = 0
    for m in range(M):
        if m == M - 1:
            deduct = 1
        for n in range(N - deduct):
            nm = m * N + n
            A[M * N + nm, n] = 6.0
            A[M * N + nm, N + nm * 2] = -2 * T[nm] ** 2
            A[M * N + nm, N + (nm + 1) * 2] = -1 * T[nm] ** 2
            A[M * N + nm, N + nm * 2 + 1] = -6 * T[nm]

    # Initialize prior vector
    m0 = np.zeros((N + 2 * M * N, 1))

    Sigma0 = np.identity(N + 2 * M * N) * sigma**2

    if len(e) == 0:
        eprior = np.ones(N) * 2 * np.pi / N
    else:
        eprior = np.array(e) * 1.0
    m0[:N, 0] = eprior * 1.0
    for m in range(M):
        for n in range(N):
            nm = m * N + n
            m0[N + nm * 2 + 1, 0] = m0[n, 0] / T[nm]

    SigmaN = Sigma0 + beta * A.T.dot(A)
    BBayes = Sigma0.dot(m0) + beta * A.T.dot(B)
    mN = np.array([spsolve(csr_matrix(SigmaN), csr_matrix(BBayes))]).T
    # Normalize encoder increments to add up to 2 pi
    epost = mN[:N, 0] * 2 * np.pi / (np.sum(mN[:N, 0]))
    return epost

def determine_mpr_speed_for_zero_crossings(
        arr_toas: np.ndarray,
        N : int, 
        M : int, 
        beta = 1E10, 
        sigma = 10
) -> pl.DataFrame:
    """ This function is used to calibrate the encoder geometry and
    calculate the shaft speeds of the encoder.

    Args:
        arr_toas (np.ndarray): The time of arrivals of the encoder.
        N (int): The number of sections in the encoder.
        M (int): The number of revolutions spanned by arr_toas.
        beta (float, optional): The beta value for the 
            Bayesian Geometry Compensation. Defaults to 1E10.
        sigma (float, optional): The sigma value for the 
            Bayesian Geometry Compensation. Defaults to 10. 
    
    Returns:
        pl.DataFrame: A DataFrame containing the shaft speeds of the encoder.
    """
    if arr_toas.shape[0] != N*M +1:
        raise ValueError("The length of arr_toas must be equal to N*M + 1")
    epost_BGC = perform_bayesian_geometry_compensation(arr_toas, N = N, M = M, beta = beta, sigma = sigma)
    arr_start_times = arr_toas[:-1].copy()
    arr_end_times = arr_toas[1:].copy()
    arr_section_distances = np.arange(N*M, dtype=np.float64)
    for m in range(M):
        arr_section_distances[m*N:(m+1)*N] = epost_BGC*1
    return pl.DataFrame({
        'section_start_time': arr_start_times,
        'section_end_time': arr_end_times,
        'encoder_section_rad': arr_section_distances
    })

def determine_mpr_shaft_speed(
        arr_toas : np.ndarray,
        N : int,
        M : int,
        beta = 1E10,
        sigma = 0.01,
        M_recalibrate : int = 7.76
    ) -> pl.DataFrame:
    """ This function is used to calibrate the encoder geometry and
    calculate the shaft speeds of the encoder.

    Args:
        arr_toas (np.ndarray): The time of arrivals of the encoder.
        N (int): The number of sections in the encoder.
        M (int): The number of revolutions spanned by arr_toas.
        beta (float, optional): The beta value for the 
            Bayesian Geometry Compensation. Defaults to 1E10.
        sigma (float, optional): The sigma value for the
            Bayesian Geometry Compensation. Defaults to 0.01.
        M_recalibrate (float, optional): The number of revolutions
            after which the encoder should be recalibrated. Defaults to 7.76.

    Returns:
        pl.DataFrame: A DataFrame containing the shaft speeds of the encoder.
    """
    len_t = len(arr_toas)
    recalibrate_interval = int(M_recalibrate*N) 

    section_measurements = {}
    for i_start in tqdm(np.array(range(0, len_t, recalibrate_interval))):
        i_end = i_start + N*M + 1
        if i_end > len_t:
            break
        df_cal = determine_mpr_speed_for_zero_crossings(
            arr_toas[i_start:i_end], 
            N = N, 
            M = M, 
            beta = beta, 
            sigma = sigma
        )
        for row_cal in df_cal.iter_rows(named=True):
            key = (row_cal["section_start_time"], row_cal["section_end_time"])
            if key in section_measurements:
                section_measurements[key].append(row_cal["encoder_section_rad"])
            else:
                section_measurements[key] = [row_cal["encoder_section_rad"]]
    
    df_section_measurements = pl.DataFrame(
        {
            'section_start_time' : [key[0] for key in section_measurements.keys()],
            'section_end_time' : [key[1] for key in section_measurements.keys()],
            'encoder_section_rads' : [val for val in section_measurements.values()]
        },
        schema = {
            "section_start_time": pl.Float64,
            "section_end_time": pl.Float64,
            "encoder_section_rads": pl.List(pl.Float64)
        }
    )
    return df_section_measurements.with_columns(
            pl.col("encoder_section_rads").list.median().alias("section_distance")
        ).with_columns(
            [
                (
                    pl.col("section_distance") / (pl.col("section_end_time") - pl.col("section_start_time"))
                ).alias("Omega")
            ]
        ).select(pl.exclude("encoder_section_rads"))

def get_mpr_geometry(df_mpr_speed : pl.DataFrame, N : int) -> pl.DataFrame:
    """ This function calculates the geometry of the MPR encoder. 

    Args:
        df_mpr_speed (pl.DataFrame): The DataFrame containing the blade AoAs. This is the
            result from the determine_mpr_shaft_speed function.
        N (int): The number of sections in the encoder.

    Returns:
        pl.DataFrame: The DataFrame containing the geometry 
            of the MPR encoder.
    """
    n_samples_total = df_mpr_speed.height
    n_random_samples = min(int(n_samples_total/10), 50)
    random_alignment_start_points = (
        df_mpr_speed
        .select(pl.arange(10*N, pl.count("Omega") - 10*N))
        .sample(n_random_samples)
    )

    df_encoder_order = None
    for n, start in enumerate(random_alignment_start_points[:, 0]):
        df_one_section = (
            df_mpr_speed[start:start+N]
            .select("section_distance")
            .with_columns(
                [
                    pl.col("section_distance"),
                    (
                        pl.col("section_distance") - pl.col("section_distance").median()
                    ).abs().alias("max_abs_section")
                ]
            ).with_columns(
                [
                    pl.when(
                        pl.col("max_abs_section") == pl.col("max_abs_section").max()
                    ).then(
                        pl.lit(1)
                    ).otherwise(
                        pl.lit(None)
                    ).alias("is_max_asb_section").fill_null(strategy="forward")
                ]
            ).with_columns(
                (
                    (
                        pl.arange(
                                pl.col("is_max_asb_section").cum_sum().max(), 
                                pl.col("is_max_asb_section").cum_sum().max() + N
                            )
                    ) % N
                ).alias("section_no")
            )
        ).sort("section_no")
        
        if df_encoder_order is None:
            df_encoder_order = df_one_section.select(
                pl.col("section_distance")
                .alias(f"section_distance_{n}")
            )
        else:
            df_encoder_order = (
                df_encoder_order
                .hstack(
                    df_one_section.select(
                        pl.col("section_distance")
                        .alias(f"section_distance_{n}")
                    )
                )
            )
    df_geometry = pl.DataFrame(
        {
            "section_distance": np.median(df_encoder_order.to_numpy(), axis=1)
        }
    ).with_columns(
        pl.col("section_distance").cum_sum().alias("section_end"),
    ).with_columns(
        (
            (
                pl.col("section_end") / pl.col("section_end").max()
            )*(2*np.pi)
        ).alias("section_end")
    ).with_columns(
        (pl.col("section_end") - pl.col("section_distance")).alias("section_start")
    ).select(
        "section_start", "section_end", "section_distance"
    )
    df_geometry[0, 'section_start'] = 0
    return df_geometry

@njit
def perform_alignment_err(
    arr_sections : np.array, 
    arr_geometry: np.array,
    arr_geometry_start: np.array,
    arr_geometry_end: np.array,
) -> np.ndarray:
    """ Get an array of alignment errors between the MPR encoder 
        and the geometry.

        This will be used to determine the best alignment between the
        start of each revolution.

    Args:
        arr_sections (np.array): The MPR sections array. This is the 
            'section_distance' column from the determine_mpr_shaft_speed
            function.
        arr_geometry (np.array): The geometry array. This is the 
            'section_distance' column from the determine_geometry function. 

    Returns:
        np.ndarray: The error between the two arrays. Must have
            the same shape as arr_sections.
    """
    arr_errors = np.ones_like(arr_sections) * -1
    N = len(arr_geometry)
    for i in np.arange(arr_sections.shape[0] - N):
        arr_errors[i] = np.sum(np.abs(arr_sections[i:i+N] - arr_geometry))
    arr_err_threshold = np.median(arr_errors) * 0.25
    arr_is_new_revo = np.zeros_like(arr_sections, dtype=np.int8)
    arr_sections_start = np.ones_like(arr_sections) * -1
    arr_sections_end = np.ones_like(arr_sections) * -1
    for i in np.arange(arr_sections.shape[0] - N):
        if arr_errors[i] < arr_err_threshold and arr_errors[i] !=-1:
            arr_is_new_revo[i] = 1
            arr_sections_start[i:i+N] = arr_geometry_start
            arr_sections_end[i:i+N] = arr_geometry_end
    return arr_is_new_revo, arr_sections_start, arr_sections_end


from typing import Literal


DATAFRAME_LIBRARY_PREFERENCE :Literal["pd", "pl"] = 'pd' 

def set_dataframe_library_preference(library: Literal["pd", "pl"]) -> None:
    """ Set the library preference for the DataFrame library.

    Args:
        library (DATAFRAME_LIBRARY_PREFERENCE): The library to use.
            Either "pd" for pandas or "pl" for polars.
    """
    if library not in ["pd", "pl"]:
        raise ValueError("The library preference must be either 'pd' or 'pl'.")
    globals()["DATAFRAME_LIBRARY_PREFERENCE"] = library


def _get_dataframe_library_preference() -> Literal["pd", "pl"]:
    """ Get the library preference for the DataFrame library.

    Returns:
        DATAFRAME_LIBRARY_PREFERENCE: The library preference.
    """
    return globals()["DATAFRAME_LIBRARY_PREFERENCE"]


from bladesight.btt import triggering_criteria
import numpy as np
import pytest

@pytest.fixture
def arr_simple_ramp_up():
    """ Simplest ramp up signal.
        (n, n) -> (n, n) for n in [0, 10)
    Returns:
        np.ndarray: A simple ramp up signal.
        np.ndarray: The time array.
    """
    return np.arange(10), np.arange(10)

@pytest.fixture
def arr_simple_ramp_down():
    """ Simplest ramp down signal.
        (n, n) -> (n, n) for n in [0, 10)

    Returns:
        np.ndarray: A simple ramp up signal.
        np.ndarray: The time array.
    """
    return np.arange(10), np.array(list(np.arange(10))[::-1])


def test_threshold_crossing_interp(arr_simple_ramp_up):
    arr_t, arr_s = arr_simple_ramp_up
    arr_toa = triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 5)
    assert len(arr_toa) == 1
    assert arr_toa[0] == 5.0
    arr_toa = triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 4.9)
    assert len(arr_toa) == 1
    assert arr_toa[0] == 4.9
    arr_toa = triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 5.1)
    assert len(arr_toa) == 1
    assert arr_toa[0] == 5.1

def test_threshold_crossing_interp_simple_first_sample(arr_simple_ramp_up):
    arr_t, arr_s = arr_simple_ramp_up
    arr_toa = triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 0)
    assert len(arr_toa) == 0
    arr_toa = triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 0.1)
    assert len(arr_toa) == 1
    assert arr_toa[0] == 0.1
    arr_toa = triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 9)
    assert len(arr_toa) == 1
    assert arr_toa[0] == 9.0
    arr_toa = triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 9.1)
    assert len(arr_toa) == 0
    # Test that triggering on the falling edge does not work
    arr_toa = triggering_criteria.threshold_crossing_interp(
        arr_t, 
        arr_s, 
        5, 
        trigger_on_rising_edge=False
    )
    assert len(arr_toa) == 0

def test_threshold_crossing_interp_falling(arr_simple_ramp_down):
    arr_t, arr_s = arr_simple_ramp_down
    arr_toa = triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 5, trigger_on_rising_edge=False)
    assert len(arr_toa) == 1
    assert np.allclose(arr_toa[0], 4.0)
    arr_toa = triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 4.9, trigger_on_rising_edge=False)
    assert len(arr_toa) == 1
    assert np.allclose(arr_toa[0],4.1)
    arr_toa = triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 5.1, trigger_on_rising_edge=False)
    assert len(arr_toa) == 1
    assert np.allclose(arr_toa[0], 3.9)

def test_threshold_crossing_interp_simple_first_sample_falling(arr_simple_ramp_down):
    arr_t, arr_s = arr_simple_ramp_down
    arr_toa = triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 0, trigger_on_rising_edge=False)
    assert len(arr_toa) == 1
    assert np.allclose(arr_toa[0], 9.0)
    arr_toa = triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 0.1, trigger_on_rising_edge=False)
    assert len(arr_toa) == 1
    assert np.allclose(arr_toa[0], 8.9)
    arr_toa = triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 9, trigger_on_rising_edge=False)
    assert len(arr_toa) == 0
    arr_toa = triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 9.1, trigger_on_rising_edge=False)
    assert len(arr_toa) == 0
    # Test that triggering on the rising edge does not work
    arr_toa = triggering_criteria.threshold_crossing_interp(
        arr_t, 
        arr_s, 
        5, 
        trigger_on_rising_edge=True
    )
    assert len(arr_toa) == 0

@pytest.fixture
def arr_zig_zag():
    """ A zig zag signal that is supposed to trigger multiple times
    when using the simple threshold.

    Returns:
        np.ndarray: A zig zag signal.
        np.ndarray: The time array.
    """
    return np.arange(10), np.array([0, 5, 0, 5, 0, 5, 0, 5, 0, 5])

def test_threshold_crossing_interp_zig_zag_rising(arr_zig_zag):
    arr_t, arr_s = arr_zig_zag
    arr_toa = triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 2.5)
    assert len(arr_toa) == 5
    assert np.allclose(arr_toa, [0.5, 2.5, 4.5, 6.5, 8.5])

def test_threshold_crossing_interp_zig_zag_falling(arr_zig_zag):
    arr_t, arr_s = arr_zig_zag
    arr_toa = triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 2.5, trigger_on_rising_edge=False)
    assert len(arr_toa) == 4
    assert np.allclose(arr_toa, [1.5, 3.5, 5.5, 7.5])

def test_wrong_n_est(arr_zig_zag):
    arr_t, arr_s = arr_zig_zag
    with pytest.raises(ValueError):
        triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 2.5, n_est=1)
    with pytest.raises(ValueError):
        triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 2.5, n_est=4)
    arr_toas_rising = triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 2.5, n_est=5)
    assert len(arr_toas_rising) == 5
    assert np.allclose(arr_toas_rising, [0.5, 2.5, 4.5, 6.5, 8.5])

    with pytest.raises(ValueError):
        triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 2.5, n_est=1, trigger_on_rising_edge=False)
    with pytest.raises(ValueError):
        triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 2.5, n_est=3, trigger_on_rising_edge=False)
    arr_toas_falling = triggering_criteria.threshold_crossing_interp(arr_t, arr_s, 2.5, n_est=4, trigger_on_rising_edge=False)
    assert len(arr_toas_falling) == 4
    assert np.allclose(arr_toas_falling, [1.5, 3.5, 5.5, 7.5])

def test_hysteresis_rising_on_simple_ramp(arr_simple_ramp_up):
    arr_t, arr_s = arr_simple_ramp_up
    arr_toas = triggering_criteria.threshold_crossing_hysteresis_rising(
        arr_t,
        arr_s,
        5,
        2,
        1
    )
    assert len(arr_toas) == 1
    assert np.allclose(arr_toas, 5.0)
    

def test_hysteresis_falling_on_simple_ramp(arr_simple_ramp_up):
    arr_t, arr_s = arr_simple_ramp_up
    arr_toas = triggering_criteria.threshold_crossing_hysteresis_falling(
        arr_t,
        arr_s,
        5,
        2,
        1
    )
    assert len(arr_toas) == 0

def test_hysteresis_rising_zig_zag_rising(arr_zig_zag):
    arr_t, arr_s = arr_zig_zag
    arr_toas = triggering_criteria.threshold_crossing_hysteresis_rising(
        arr_t,
        arr_s,
        2.5,
        1,
        10
    )
    assert len(arr_toas) == 5
    assert np.allclose(arr_toas, [0.5, 2.5, 4.5, 6.5, 8.5])

def test_hysteresis_rising_zig_zag_falling(arr_zig_zag):
    arr_t, arr_s = arr_zig_zag
    arr_toas = triggering_criteria.threshold_crossing_hysteresis_falling(
        arr_t,
        arr_s,
        2.5,
        1,
        10
    )
    assert len(arr_toas) == 4
    assert np.allclose(arr_toas, [1.5, 3.5, 5.5, 7.5])

@pytest.fixture
def arr_zig_zag_multiple():
    """ A zig zag signal that is supposed to trigger multiple times
    when using the simple threshold.

    Returns:
        np.ndarray: A zig zag signal.
        np.ndarray: The time array.
    """
    return np.arange(10), np.array([0, 10, 5, 10, 5, 10, 0, 10, 5, 10])

def test_hysteresis_rising_zig_zag_multiple_rising(arr_zig_zag_multiple):
    arr_t, arr_s = arr_zig_zag_multiple
    arr_toas = triggering_criteria.threshold_crossing_hysteresis_rising(
        arr_t,
        arr_s,
        10,
        5
    )
    assert len(arr_toas) == 5
    assert np.allclose(arr_toas, [1, 3, 5, 7, 9])
    arr_toas = triggering_criteria.threshold_crossing_hysteresis_rising(
        arr_t,
        arr_s,
        10,
        5.01
    )
    assert len(arr_toas) == 2
    assert np.allclose(arr_toas, [1, 7])

def test_hysteresis_rising_zig_zag_multiple_falling(arr_zig_zag_multiple):
    arr_t, arr_s = arr_zig_zag_multiple
    arr_toas = triggering_criteria.threshold_crossing_hysteresis_falling(
        arr_t,
        arr_s,
        5,
        5
    )
    assert len(arr_toas) == 4
    assert np.allclose(arr_toas, [2, 4, 5.5, 8])
    arr_toas = triggering_criteria.threshold_crossing_hysteresis_falling(
        arr_t,
        arr_s,
        5,
        5.01
    )
    assert len(arr_toas) == 0



# Test threshold_crossing_interp todo:
# - Test with n_est, ensure exception is raised if 
#    n_est is less than the number of ToAs
# - Test with falling edge triggering
# - Test with n_est and falling edge triggering
# - ...

    



import bladesight.dataset_handler
import os

def test_online_dataset_lookup_failing(monkeypatch):    
    def mock_get_bladesight_datasets():
        return []

    monkeypatch.setattr(bladesight.dataset_handler, "get_bladesight_datasets", mock_get_bladesight_datasets)
    Datasets = bladesight.dataset_handler.BladesightDatasetDirectory()
    assert Datasets.online_datasets == []
    Datasets._ipython_key_completions_()
    Datasets.local_datasets = []
    assert Datasets._ipython_key_completions_() == []

    def mock_get_bladesight_datasets():
        return [
            "bladesight-data-mock/test1",
            "bladesight-data-mock/test2",
            "bladesight-data-mock/test3"
        ]
    monkeypatch.setattr(bladesight.dataset_handler, "get_bladesight_datasets", mock_get_bladesight_datasets)
    Datasets = bladesight.dataset_handler.BladesightDatasetDirectory()
    Datasets._ipython_key_completions_()
    assert Datasets.online_datasets == [
        "bladesight-data-mock/test1",
        "bladesight-data-mock/test2",
        "bladesight-data-mock/test3"
    ]
    
    Datasets.local_datasets = []
    assert Datasets._ipython_key_completions_() == [
        "data/test1",
        "data/test2",
        "data/test3"
    ]
    
def test_default_to_local_datasets(tmp_path, monkeypatch):
    def mock_get_bladesight_datasets():
        raise ValueError("Mocked error")
    monkeypatch.setattr(bladesight.dataset_handler, "get_bladesight_datasets", mock_get_bladesight_datasets)
    os.environ["BLADESIGHT_DATASETS_PATH"] = str(tmp_path)
    path_to_bladesight = bladesight.dataset_handler.get_path_to_local_bladesight()
    path_to_folder1 = path_to_bladesight / "folder1"
    path_to_folder1.mkdir(parents=True, exist_ok=True)
    Datasets = bladesight.dataset_handler.BladesightDatasetDirectory()
    assert Datasets.local_datasets == []
    assert Datasets._ipython_key_completions_() == []
    path_to_db = path_to_folder1 / "file1.db"
    path_to_csv = path_to_folder1 / "file2.csv"
    path_to_db.touch()
    path_to_csv.touch()
    Datasets = bladesight.dataset_handler.BladesightDatasetDirectory()
    assert Datasets.local_datasets == ["folder1/file1"]
    Datasets._ipython_key_completions_()
    assert Datasets.online_datasets == ["folder1/file1"]
    assert Datasets._ipython_key_completions_() == ["data/file1"]

    path_to_db = path_to_folder1 / "file2.db"
    path_to_db.touch()
    Datasets._refresh_available_datasets()
    assert Datasets.local_datasets == ["folder1/file1", "folder1/file2"]
    assert Datasets.online_datasets == ["folder1/file1", "folder1/file2"]
    assert Datasets._ipython_key_completions_() == ["data/file1", "data/file2"]


import json
import os
import pathlib
from typing import Dict, List, Literal, Union, Any, Optional
from typing_extensions import Self
import duckdb
import pandas as pd
import polars as pl
import s3fs
from yaspin import yaspin
from .utils import ( 
    _get_dataframe_library_preference
)

BLADESIGHT_DATASETS_S3_BUCKET = "bladesight-datasets"

def get_path_to_local_bladesight() -> pathlib.Path:
    """This function returns the path to the local datasets folder.
    If there is no environmental variable called BLADESIGHT_DATASETS_PATH, it
    will return ~/.bladesight.

    Returns:
        pathlib.Path: The path to the local datasets folder. It does
            not necessarily exist.
    """
    if "BLADESIGHT_DATASETS_PATH" in os.environ:
        return pathlib.Path(os.environ["BLADESIGHT_DATASETS_PATH"]) / ".bladesight"
    else:
        return pathlib.Path.home() / ".bladesight"

def get_local_datasets() -> List[str]:
    """This function returns a list of the names of the datasets in the local
    datasets folder.

    Returns:
        List[str]: A list of the dataset names in the in the 
            local datasets folder.
    """
    BLADESIGHT_DATASETS_PATH = get_path_to_local_bladesight()
    if not BLADESIGHT_DATASETS_PATH.exists():
        return []
    else:
        local_datasets = []
        for path_root, _, files in os.walk(BLADESIGHT_DATASETS_PATH):
            for file in files:
                if file.endswith(".db"):
                    path_parts = pathlib.Path(path_root).parts
                    path_prefix = None
                    add_parts = False
                    for part in path_parts:
                        if add_parts:
                            if path_prefix is None:
                                path_prefix = part
                            else:
                                path_prefix = path_prefix + "/" + part
                        if part == ".bladesight":
                            add_parts = True
                    local_datasets.append(f"{path_prefix}/{file}"[:-3])
        return local_datasets

def get_bladesight_datasets() -> List[str]:
    """This function returns a list of all the datasets in
        the bladesight-datasets bucket.

    Returns:
        List[str]: A list of the names of the datasets in the bucket.
    """
    s3 = s3fs.S3FileSystem(anon=True)
    datasets = []
    with yaspin(text="Getting all datasets from Bladesight Data..."):
        for bucket_root, _, files in s3.walk(BLADESIGHT_DATASETS_S3_BUCKET + "/"):
            for file in files:
                if file.endswith(".db"):
                    datasets.append(f"{bucket_root}/{file}"[:-3])
    return datasets

def download_dataset_from_bladesight_data(dataset_path_on_s3: str) -> None:
    """This function downloads a dataset from S3 and saves it locally.
    
    Args:
        dataset_path_on_s3 (str): The path to the dataset on S3.

    Examples:
    ---------
        Download a dataset into the local .bladesight directory

        >>> download_dataset_from_bladesight_data("bladesight-datasets/intro_to_btt/intro_to_btt_ch02")
    """
    s3 = s3fs.S3FileSystem(anon=True)
    PATH_TO_LOCAL_DB = get_path_to_local_bladesight() / "data"
    for s3_subfolder in dataset_path_on_s3.split("/")[1:]:
        PATH_TO_LOCAL_DB = PATH_TO_LOCAL_DB / s3_subfolder

    if not PATH_TO_LOCAL_DB.parent.exists():
        PATH_TO_LOCAL_DB.parent.mkdir(parents=True)

    with yaspin(
        text=f"Downloading {dataset_path_on_s3} from Bladesight Data..."
    ) as spinner:
        s3.download(dataset_path_on_s3 + ".db", str(PATH_TO_LOCAL_DB) + ".db")
    spinner.text = f"Done downloading {dataset_path_on_s3} from Bladesight Data... "
    spinner.ok("✅ ")

def _confirm_dataset_is_valid(path_to_db : pathlib.Path) -> None:
    """This function checks if a dataset exists 
    and has a .db extension in its name.

    Args:
        path_to_db (pathlib.Path): The path to the dataset.

    Raises:
        FileNotFoundError: If the dataset does not exist.
        ValueError: If the dataset does not have a .db extension.

    Examples:
    ---------
        Check that a dataset has a .db extention.
        
        >>> _confirm_dataset_is_valid("bladesight-data/intro_to_btt/intro_to_btt_ch02.db")
    """
    if not path_to_db.exists():
        raise FileNotFoundError(
            f"You are trying to open the dataset {path_to_db},"
            " but it does not exist!"
        )
    if path_to_db.suffix != ".db":
        raise ValueError(
            f"You are trying to open the dataset {path_to_db}, "
            "but it does not have a .db extension!"
        )

def _execute_sql_without_arg(
    path_to_db : pathlib.Path, 
    sql_query: str, 
    return_mode: Literal['pd', 'pl', 'no_return'] = 'pd'
) -> Union[pd.DataFrame, pl.DataFrame, None]:
    """This function executes a DuckDB SQL query on the dataset
    and returns its result as a pandas or polars DataFrame.

    Args:
        path_to_db (pathlib.Path): The path to the dataset.
        sql_query (str): The SQL query to execute.
        return_mode (Literal['pd', 'pl', 'no_return'], optional): The return 
            mode. Defaults to 'pd'.

    Returns:
        Union[pd.DataFrame, pl.DataFrame, None]: The results of the query.

    Raises:
        ValueError: If the return_mode is not 'pd', 'pl' or 'no_return'.
    
    Examples:
    ---------
        Read a SQL query from a dataset into a Polars DataFrame.

        >>> _execute_sql_without_arg(
        ... "bladesight-data/intro_to_btt/intro_to_btt_ch02.db", 
        ... "SELECT * FROM metadata;",
        ... "pl"
        ... )
    """
    if return_mode not in ["pd", "pl", "no_return"]:
        raise ValueError("return_mode must be 'pd', 'pl' or 'no_return'")
    
    with duckdb.connect(str(path_to_db)) as con:
        if return_mode == "pd":
            df = con.execute(sql_query).df()
        elif return_mode == "pl":
            df = con.execute(sql_query).pl()
        elif return_mode == "no_return":
            con.execute(sql_query)
            df = None
    return df

def _execute_sql_with_arg(
    path_to_db : pathlib.Path, 
    sql_query: str, 
    df_in_memory: Union[pd.DataFrame, pl.DataFrame],
    return_mode: Literal["pd", "pl", 'no_return'] = "pd"
) -> Union[pd.DataFrame, pl.DataFrame, None]:
    """This function executes a DuckDB SQL query on the dataset
    and returns its result as a pandas or polars DataFrame.

    Args:
        path_to_db (pathlib.Path): The path to the dataset.
        sql_query (str): The SQL query to execute.
        df_in_memory (Union[pd.DataFrame, pl.DataFrame]): The DataFrame to use 
            in the query.
        return_mode (Literal["pd", 'pl', 'no_return'], optional): The return 
            mode. Defaults to 'pd'.

    Returns:
        Union[pd.DataFrame, pl.DataFrame, None]: The results of the query.

    Raises:
        ValueError: If the return_mode is not 'pd', 'pl' or 'no_return'.
    
    Examples:
    ---------
        Read a SQL query from a dataset into a Polars DataFrame.

        >>> _execute_sql_without_arg(
        ... "bladesight-data/intro_to_btt/intro_to_btt_ch02.db", 
        ... "SELECT * FROM metadata;",
        ... "pl"
        ... )
    """
    if return_mode not in ["pd", "pl", "no_return"]:
        raise ValueError("return_mode must be 'pd', 'pl' or 'no_return'")
    
    with duckdb.connect(str(path_to_db)) as con:
        if return_mode == "pd":
            df = con.execute(sql_query).df()
        elif return_mode == "pl":
            df = con.execute(sql_query).pl()
        elif return_mode == "no_return":
            con.execute(sql_query)
            df = None
    return df

def _set_metadata_key(
        path_to_db : pathlib.Path, 
        metadata_key : str, 
        metadata: Dict[str, Dict]
    ) -> None:
    """This function sets the metadata in the dataset.

    Args:
        path_to_db (pathlib.Path): The path to the dataset.
        metadata_key (str): The key for the metadata.
        metadata (Dict[str, Dict]): The metadata.

    Examples:
    ---------
        Set the metadata in a dataset.

        >>> _set_metadata(
        ... "bladesight-data/intro_to_btt/intro_to_btt_ch02.db",
        ... {
        ...     "CITATION": {
        ...         "repr": "This is a citation",
        ...         "url": "https://example.com",
        ...         "doi": "10.1234/5678"
        ...     }
        ... }
        ... )
    """
    with duckdb.connect(str(path_to_db)) as con:
        con.execute("DELETE FROM metadata WHERE metadata_key = ?;", (metadata_key, ))
        metadata_value_encoded = json.dumps(metadata)
        con.execute(
            "INSERT INTO metadata VALUES (?, ?);",
            (
                metadata_key,
                metadata_value_encoded,
            ),
        )

def _get_all_metadata(path_to_db : pathlib.Path) -> Dict[str, Union[Dict, Any]]:
    """This function returns a metadata dictionary
    from the metadata table in the dataset.

    Args:
        path_to_db (pathlib.Path): The path to the dataset.

    Returns:
        Dict[str, Union[Dict, Any]]: The metadata.

    Examples:
    ---------
        Get all the metadata from a dataset.

        >>> _get_all_metadata("bladesight-data/intro_to_btt/intro_to_btt_ch02.db")
        {
            "CITATION": {
                "repr": "This is a citation",
                "url": "https://example.com",
                "doi": "10.1234/5678"
            }
        }
    """
    df_metadata = _execute_sql_without_arg(path_to_db, "SELECT * FROM metadata;")
    metadata = {}
    for _, row in df_metadata.iterrows():
        metadata[row["metadata_key"]] = json.loads(row["metadata_value"])
    return metadata

def _get_db_tables(path_to_db : pathlib.Path) -> List[str]:
    """This method gets the tables in the dataset. It
    excludes the metadata table.

    Args:
        path_to_db (pathlib.Path): The path to the dataset.

    Returns:
        List[str]: The tables in the dataset.

    Examples:
    ---------
        Get all the tables in a dataset.

        >>> _get_db_tables("bladesight-data/intro_to_btt/intro_to_btt_ch02.db")
        ['dataset_1', 'dataset_2']
    """
    all_tables = _execute_sql_without_arg(
        path_to_db,
        "SHOW TABLES;"
    )["name"].to_list()
    data_tables = list(set(all_tables) - set(["metadata"]))
    return sorted(data_tables)

def _get_printable_citation(metadata: Dict[str, Dict]) -> str:
    """This function returns a printable citation 
    from the metadata.

    Args:
        metadata (Dict[str, Dict]): The metadata.

    Returns:
        str: The printable citation.

    Examples:
    ---------
        Get a printable citation from the metadata.

        >>> _get_printable_citation({
        ... "CITATION": {
        ...     "repr": "This is a citation",
        ...     "url": "https://example.com",
        ...     "doi": "10.1234/5678"
        ...     }
        ... })
        "If you use this dataset in published work, please \
         use the below citation:\n\nThis is a citation\nLink\
         to paper: https://example.com\nDOI: 10.1234/5678"
    """
    if "CITATION" not in metadata.keys():
        return "No citation provided in metadata table."
    
    citation = metadata["CITATION"]
    if not isinstance(citation, dict):
        return "Citation format error. Here is the raw citation:\n\n" + str(citation)
    
    if "repr" not in citation.keys():
        return "Citation format error. Here is the raw citation:\n\n" + str(citation)

    cite_main = f"""If you use this dataset in published work, please use the below citation:\n\n{citation['repr']}"""

    if "url" in citation.keys():
        cite_main += f"""\nLink to paper: {citation['url']}"""

    if "doi" in citation.keys():
        cite_main += f"""\nDOI: {citation['doi']}"""

    if "license" in citation.keys():
        cite_main += f"""\nLicense: {citation['license']}"""
    
    return cite_main

class Dataset:
    """This object is used to access data from a dataset.
    
    Args:
        path (pathlib.Path): The path to the dataset.
    
    Examples:
    ---------
        >>> dataset = Dataset("bladesight-data/intro_to_btt/intro_to_btt_ch02.db")
        >>> dataset.tables
        ['dataset_1', 'dataset_2']
        >>> dataset.metadata
        {
            "CITATION": {
                "repr": "This is a citation",
                "url": "https://example.com",
                "doi": "10.1234/5678"
            }
        }
        >>> df_table = dataset["table/dataset_1"]
        >>> dataset.print_citation()
    """
    def __init__(self, path: pathlib.Path):
        _confirm_dataset_is_valid(path)
        self.path = path
        self.tables: List[str] = _get_db_tables(self.path)
        self.metadata: Dict[str, Dict] = _get_all_metadata(self.path)
        self.print_citation()
    
    # Create a getter and setter for the metadata.doi property
    @property
    def doi(self) -> str:
        """This function returns the DOI from the metadata.

        Returns:
            str: The DOI.

        Examples:
        ---------
            Get the DOI from the metadata.

            >>> dataset = Dataset("bladesight-data/intro_to_btt/intro_to_btt_ch02.db")
            >>> dataset.doi
            "10.1234/5678"
        """
        return self.metadata["CITATION"]["doi"]
    
    def _set_citation_field(
            self, 
            metadata_key : Literal["repr", "url", "doi"], 
            new_value : str
        ):
        """This function sets a metadata field in the dataset.

        Args:
            metadata_key (Literal["repr", "url", "doi"]): The metadata key.
            new_value (str): The new value.
        
        Examples:
        ---------
            Set the metadata field in the dataset.

            >>> dataset = Dataset("bladesight-data/intro_to_btt/intro_to_btt_ch02.db")
            >>> dataset._set_citation_field("doi", "10.1234/5678")
        """
        self.metadata["CITATION"][metadata_key] = new_value
        _set_metadata_key(self.path, "CITATION", self.metadata['CITATION'])
        self.metadata: Dict[str, Dict] = _get_all_metadata(self.path)

    @doi.setter
    def doi(self, new_doi: str):
        """This function sets the DOI in the metadata.

        Args:
            new_doi (str): The new DOI.

        Examples:
        ---------
            Set the DOI in the metadata.

            >>> dataset = Dataset("bladesight-data/intro_to_btt/intro_to_btt_ch02.db")
            >>> dataset.doi = "10.1234/5678"
        """
        self._set_citation_field("doi", new_doi)

    @property
    def url(self) -> str:
        """This function returns the URL from the metadata.

        Returns:
            str: The URL.

        Examples:
        ---------
            Get the URL from the metadata.

            >>> dataset = Dataset("bladesight-data/intro_to_btt/intro_to_btt_ch02.db")
            >>> dataset.url
            "https://example.com"
        """
        return self.metadata["CITATION"]["url"]
    
    @url.setter
    def url(self, new_url: str):
        """This function sets the URL in the metadata.

        Args:
            new_url (str): The new URL.

        Examples:
        ---------
            Set the URL in the metadata.

            >>> dataset = Dataset("bladesight-data/intro_to_btt/intro_to_btt_ch02.db")
            >>> dataset.url = "https://example.com"
        """
        self._set_citation_field("url", new_url)

    @property
    def citation(self) -> str:
        """This function returns the citation from the metadata.
        This is syntactic sugar for the repr field in the CITATION.

        Returns:
            str: The citation.

        Examples:
        ---------
            Get the citation from the metadata.

            >>> dataset = Dataset("bladesight-data/intro_to_btt/intro_to_btt_ch02.db")
            >>> dataset.citation
            "This is a citation\nLink to paper: https://example.com\nDOI: 10.1234/5678"
        """
        return self.metadata["CITATION"]["repr"]

    @citation.setter
    def citation(self, new_citation: str):
        """This function sets the citation in the metadata.
        This is syntactic sugar for the repr field in the CITATION.

        Args:
            new_citation (str): The new citation.

        Examples:
        ---------
            Set the citation in the metadata.

            >>> dataset = Dataset("bladesight-data/intro_to_btt/intro_to_btt_ch02.db")
            >>> dataset.citation = "This is a citation"
        """
        self._set_citation_field("repr", new_citation)
    

    @property
    def license(self) -> str:
        """This function returns the license from the metadata.

        Returns:
            str: The license.

        Examples:
        ---------
            Get the license from the metadata.

            >>> dataset = Dataset("bladesight-data/intro_to_btt/intro_to_btt_ch02.db")
            >>> dataset.license
            "CC BY 4.0"
        """
        return self.metadata["license"]
    
    @license.setter
    def license(self, new_license: str):
        """This function sets the license in the metadata.

        Args:
            new_license (str): The new license.

        Examples:
        ---------
            Set the license in the metadata.

            >>> dataset = Dataset("bladesight-data/intro_to_btt/intro_to_btt_ch02.db")
            >>> dataset.license = "CC BY 4.0"
        """
        self._set_citation_field("license", new_license)

    def query(
            self, 
            sql_query: str, 
            df_in_memory : Optional[Union[pd.DataFrame, pl.DataFrame]] = None,
            no_return : bool = False
        ) -> Union[pd.DataFrame, pl.DataFrame]:
        """This function executes a SQL query on the dataset
        and returns its result as a pandas or polars DataFrame.

        Args:
            sql_query (str): The SQL query to execute.
            df_in_memory (Optional[Union[pd.DataFrame, pl.DataFrame]], optional):
                An in-memory DataFrame to pass through to the query statement.
                Defaults to None.
            no_return (bool, optional): If True, the query will not return a DataFrame.
                Use this with custom create statements.

        Returns:
            Union[pd.DataFrame, pl.DataFrame]: The results of the query.

        Examples:
        ---------
            Query a dataset using SQL.

            >>> dataset = Dataset("bladesight-data/intro_to_btt/intro_to_btt_ch02.db")
            >>> df = dataset.query("SELECT * FROM table_1;")
        """
        if df_in_memory is not None:
            if not isinstance(df_in_memory, (pd.DataFrame, pl.DataFrame)):
                raise ValueError("df_in_memory must be a pandas or polars DataFrame")
            return _execute_sql_with_arg(
                self.path, 
                sql_query, 
                df_in_memory, 
                return_mode=_get_dataframe_library_preference() if not no_return else "no_return"
            )
        return _execute_sql_without_arg(
            self.path, 
            sql_query, 
            return_mode=_get_dataframe_library_preference() if not no_return else "no_return"
        )
    
    def __getitem__(self, key: str) -> Union[pd.DataFrame, pl.DataFrame]:
        """ This function returns a table from the dataset.
        
        Args:
            key (str): The name of the table, prefixed with "table/".

        Raises:
            KeyError: If the table is not found.

        Returns:
            Union[pd.DataFrame, pl.DataFrame]: The table.

        Examples:
        ---------
            Load a table from the dataset into memory:

            >>> dataset = Dataset("bladesight-data/intro_to_btt/intro_to_btt_ch02.db")
            >>> df_table = dataset["table/dataset_1"]
        """
        table_name = key.replace("table/", "")
        if table_name in self.tables:
            return _execute_sql_without_arg(
                self.path,
                f"SELECT * FROM {table_name};", 
                return_mode=_get_dataframe_library_preference()
            )
        else:
            raise KeyError(
                f"Table {table_name} not found. These are the tables in the dataset: {self.tables}"
            )
    
    def print_citation(self):
        """Print the citation provided in the metadata table."""
        print(_get_printable_citation(self.metadata))
    
    def _ipython_key_completions_(self):
        return ["table/" + i for i in self.tables]
    
    def __repr__(self) -> str:
        """Show the dataset and its tables.

        Returns:
            str: The dataset and its tables in a string.
        """
        table_string = "[\n"
        for table in self.tables:
            table_string += f"\t'table/{table}',\n "
        table_string += "]"
        return f"Dataset({self.path}),\n\n Tables: \n {table_string}"

    def __setitem__(self, key: str, df: Union[pd.DataFrame, pl.DataFrame]):
        """This function sets a table in the dataset.

        Args:
            key (str): The name of the table.
            df (Union[pd.DataFrame, pl.DataFrame]): The table.

        Examples:
        ---------
            Set a table in the dataset.

            >>> dataset = Dataset("bladesight-data/intro_to_btt/intro_to_btt_ch02.db")
            >>> df = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
            >>> dataset["table/my_table"] = df
        """
        if not isinstance(df, (pd.DataFrame, pl.DataFrame)):
            raise ValueError("df must be a pandas or polars DataFrame")
        table_name = key.replace("table/", "")
        if table_name in self.tables:
            raise ValueError(
                f"""Table {table_name} already exists in the dataset. """
                """\nIf you want to overwrite this table, use the .drop_table method to drop"""
                """it first."""
            )
        _execute_sql_with_arg(
            self.path,
            f"CREATE OR REPLACE TABLE {table_name} AS SELECT * FROM df",
            df,
            return_mode="no_return"
        )
        print(f"✅ Created table {table_name} in dataset {self.path}")
        self.tables: List[str] = _get_db_tables(self.path)
    
    def refresh_tables(self):
        """This function refreshes the tables in the dataset.

        Examples:
        ---------
            Refresh the tables in the dataset.

            >>> dataset = Dataset("bladesight-data/intro_to_btt/intro_to_btt_ch02.db")
            >>> dataset.refresh_tables()
        """
        self.tables: List[str] = _get_db_tables(self.path)
    
    def drop_table(self, table_name: str) -> Self:
        """This function drops a table in the dataset.

        Args:
            key (str): The name of the table.

        Examples:
        ---------
            Drop a table in the dataset.

            >>> dataset = Dataset("bladesight-data/intro_to_btt/intro_to_btt_ch02.db")
            >>> dataset.drop_table("table/my_table")
        """
        table_name = table_name.replace("table/", "")
        if table_name in self.tables:
            _execute_sql_without_arg(
                self.path,
                f"DROP TABLE {table_name};",
                return_mode="no_return"
            )
            print(f"✅ Dropped table {table_name} in dataset {self.path}")
            self.tables: List[str] = _get_db_tables(self.path)
        else:
            pass
        return self

class BladesightDatasetDirectory:
    """This object is used to access datasets from the 
    Bladesight Data bucket on S3. 
    
    It also lists the local datasets.

    Examples:
    ---------
    Load a dataset into memory:

        >>> Datasets = BladesightDatasetDirectory()
        >>> dataset = Datasets["data/intro_to_btt/intro_to_btt_ch02"]
        >>> df_table = dataset["table/dataset_1"]
    """
    def __init__(self):
        self.path = get_path_to_local_bladesight()
        self.local_datasets = [
            self._replace_path_prefix(i) for i in get_local_datasets()
        ]
        self.local_datasets :  List[str] = get_local_datasets()
        self.online_datasets : List[str] = []

        self.online_loaded = False

    @staticmethod
    def _getitem_key_correct_format(key: str) -> bool:
        """This function checks if the key is in the correct format. The key
        should be in the format "data/intro_to_btt/intro_to_btt_ch02".

        Args:
            key (str): The key to check.

        Returns:
            bool: True if the key is in the correct format, False otherwise.

        Examples:
        ---------
        Check if a key is in the correct format:

            >>> BladesightDatasetDirectory._getitem_key_correct_format(
            ... "data/intro_to_btt/intro_to_btt_ch02"
            ... )
            True

            >>> BladesightDatasetDirectory._getitem_key_correct_format(
            ... "intro_to_btt/intro_to_btt_ch02"
            ... )
            False
        """
        if key.startswith("data/"):
            return True
        return False
    
    def __getitem__(self, key: str) -> Dataset:
        """Get the dataset specified by a key. If the dataset is not found, it
        will be downloaded from the Bladesight Data bucket.

        Args:
            key (str): The name of the dataset.

        Raises:
            KeyError: If the dataset is not found.

        Returns:
            Dataset: The dataset.

        Examples:
        ---------
        Load a dataset into memory:

            >>> Datasets = BladesightDatasetDirectory()
            >>> dataset = Datasets["data/intro_to_btt/intro_to_btt_ch02"]
        """
        if self._getitem_key_correct_format(key) is False:
            raise KeyError(
                f"Dataset {key} does not start with data/. The key should be in the format 'data/../../etc'."
            )
        for local_dataset in self.local_datasets:
            homogenized_local_name = self._replace_path_prefix(local_dataset)
            if key == homogenized_local_name:
                return Dataset(self.path / pathlib.Path(local_dataset + ".db"))
        else:
            # Download the dataset from the online datasets
            if self.online_loaded is False:
                self._refresh_available_datasets()
                self.online_loaded = True
            for online_set in self.online_datasets:
                homogenized_online_name = self._replace_path_prefix(online_set)
                if key == homogenized_online_name:
                    download_dataset_from_bladesight_data(
                        self._replace_path_prefix(key, BLADESIGHT_DATASETS_S3_BUCKET)
                    )
                    self.local_datasets = get_local_datasets()
                    return self[key]

            else:
                raise KeyError(f"Dataset {key} not found.")
    
    @staticmethod
    def _replace_path_prefix(
        dataset_full_path: str, replace_prefix: str = "data"
    ) -> str:
        """This function is used to replace the first path prefix with the
            replace_prefix argument. For example, if the dataset path
            is "bladesight-data/intro_to_btt/intro_to_btt_ch02", and the
            replace_prefix is "data", and the path is returned as
            "data/intro_to_btt/intro_to_btt_ch02".

        Args:
            dataset_full_path (str): The full path to the dataset.
            replace_prefix (str, optional): The prefix to replace. Defaults to "data".

        Returns:
            str: The new path.

        Examples:
        ---------
        Replace the first path prefix with "data":
        
            >>> _replace_path_prefix(
            ... "bladesight-data/intro_to_btt/intro_to_btt_ch02", 
            ... "data"
            ... )
            "data/intro_to_btt/intro_to_btt_ch02"
        """
        new_path = [replace_prefix] + dataset_full_path.split("/")[1:]
        return "/".join(new_path)
    
    def _ipython_key_completions_(self):
        """ 
        We replace whatever prefix is in the 
        dataset with "data" 
        """
        if self.online_loaded is False:
            self._refresh_available_datasets()
            self.online_loaded = True
        local_sets_homogenized = [self._replace_path_prefix(i) for i in self.local_datasets]
        online_sets_homogenized = [self._replace_path_prefix(i) for i in self.online_datasets]
        composite_set = local_sets_homogenized
        for i in online_sets_homogenized:
            if i not in composite_set:
                composite_set.append(i)
        return composite_set
    
    def _refresh_available_datasets(self):
        """
        This function refreshes the local and online datasets.
        If the online datasets cannot be read, it will only 
        list the local datasets.
        """
        self.local_datasets = get_local_datasets()
        try:
            self.online_datasets = get_bladesight_datasets()
        except Exception as _:
            print("Could not read remote datasets. Only listing local datasets")
            self.online_datasets = self.local_datasets

    def new_dataset(
            self, 
            dataset_name : str, 
            exist_ok : bool = False 
        ) -> Dataset:
        """This function creates a new dataset object.

        Args:
            dataset_name (str): The name of the dataset.
            exist_ok (bool, optional): If True, the function will not raise an error
                if the dataset already exists. Defaults to False.
        
        Returns:
            Dataset: The dataset object.

        Examples:
        ---------
        Create a new dataset object:

            >>> Datasets = BladesightDatasetDirectory()
            >>> dataset = Datasets.new_dataset("intro_to_btt/intro_to_btt_ch02_my_dev")
        """
        if not dataset_name.startswith("data/"):
            add_prefix = "data/"
        else:
            add_prefix = ""
        new_dataset_path = self.path / pathlib.Path(add_prefix + dataset_name + ".db")
        
        # Check if the dataset already exists
        if new_dataset_path.exists():
            if exist_ok:
                return self[add_prefix + dataset_name]
            else:
                raise FileExistsError(f"Dataset {dataset_name} already exists.")
        
        # Create a new dataset
        pathlib.Path(new_dataset_path.parent).mkdir(parents=True, exist_ok=True)
        with duckdb.connect(str(new_dataset_path)) as con:
            con.execute("CREATE TABLE metadata (metadata_key TEXT, metadata_value TEXT);")
        init_citation = {
                "url" : "No citation provided in metadata table.",
                "doi" : "No citation provided in metadata table.",
                "repr": "No citation provided in metadata table."
        }
        _set_metadata_key(new_dataset_path, "CITATION", init_citation)
        ds = Dataset(new_dataset_path)
        self.local_datasets = get_local_datasets()
        return ds

Datasets = BladesightDatasetDirectory()



from typing import Union
import polars as pl
import pandas as pd
import numpy as np
from .geometry_compensation import (
    determine_mpr_shaft_speed, 
    get_mpr_geometry, 
    perform_alignment_err
)
from ..utils import _get_dataframe_library_preference
import duckdb

def calculate_ias(
        arr_toas: Union[np.ndarray, pd.DataFrame, pl.DataFrame],
        N: int, 
        M: int = 10, 
        beta : float = 1E10,
        sigma : float = 0.01,
        M_recalibrate : float = 7.76,
    ) -> Union[pl.DataFrame ,pd.DataFrame]:
    """ Calculate the shaft speed and corresponding sections
        of the MPR encoder.

    Args:
        arr_toas (Union[np.ndarray, pd.DataFrame, pl.DataFrame]): The time of 
            arrivals of the encoder. If a Pandas or Polars DataFrame is
            supplied, the first column will be converted to a numpy array.
        N (int): The number of sections in the encoder.
        M (int): The number of revolutions spanned by arr_toas. Defaults to 10.
        beta (float, optional): The beta value for the 
            Bayesian Geometry Compensation. Defaults to 1E10.
        sigma (float, optional): The sigma value for the
            Bayesian Geometry Compensation. Defaults to 0.01.
        M_recalibrate (float, optional): The number of revolutions
            after which the encoder should be recalibrated. Defaults to 7.76.

    Returns:
        pl.DataFrame | pd.DataFrame: A DataFrame containing the shaft 
            speeds of the encoder.
    """
    if M_recalibrate >= M:
        raise ValueError(
            "M_recalibrate must be less than M. Try using the default values first"
            ", they are usually good enough."
        )
    if isinstance(arr_toas, (pd.DataFrame, pl.DataFrame)):
        arr_toas = duckdb.query(f"""SELECT {list(
            duckdb.execute(
                "SELECT * FROM arr_toas LIMIT 0"
            ).fetchnumpy().keys()
        )[0]} as toa FROM arr_toas""").fetchnumpy()["toa"]
    print("Calculating the IAS. This may take a while...")
    df_mpr = determine_mpr_shaft_speed(
        arr_toas,
        N = N,
        M = M,
        beta = beta,
        sigma = sigma,
        M_recalibrate = M_recalibrate
    )
    df_geometry = get_mpr_geometry(df_mpr, N)
    arr_is_new_revo, arr_sections_start, arr_sections_end = perform_alignment_err(
        df_mpr["section_distance"].to_numpy(),
        df_geometry["section_distance"].to_numpy(),
        df_geometry["section_start"].to_numpy(),
        df_geometry["section_end"].to_numpy()
    )
    df_mpr_speed = df_mpr.with_columns(
        [
            pl.Series("n",arr_is_new_revo).cum_sum().alias("n"),
            pl.Series("section_start", arr_sections_start),
            pl.Series("section_end", arr_sections_end),
        ]
    ).filter(
        pl.col("section_start") >= 0
    )
    if _get_dataframe_library_preference() == "pd":
        return df_mpr_speed.to_pandas()
    return df_mpr_speed



from typing import List
import pandas as pd
import numpy as np

def pivot_blade_AoAs_along_revolutions(
    prox_AoA_dfs : List[pd.DataFrame]
) -> pd.DataFrame:
    """This function aligns the AoA DataFrames (from the 
        transform_prox_AoAs_to_blade_AoAs function) along
        the shaft revolutions.

        The function returns a DataFrame with the ToA
        and AoA values for each blade having a column.

    Args:
        prox_AoA_dfs (List[pd.DataFrame]): A list of DataFrames
        where each DataFrame contains the ToA and AoA values
        for a single blade from a proximity probe.

    Returns:
        pd.DataFrame: A DataFrame where every row contains the
        data pertaining to a single shaft revolution and every 
        blade's ToA and AoA values are in its own column respectively.
    """
    df_blades_aligned = prox_AoA_dfs[0]
    df_blades_aligned = df_blades_aligned.rename(columns={"ToA":"ToA_1", "AoA":"AoA_1"})
    for i, df_blade in enumerate(prox_AoA_dfs[1:]):
        df_blades_aligned = df_blades_aligned.merge(
            df_blade[["n", "ToA", "AoA"]].rename(
                columns={
                    "ToA":"ToA_"+str(i+2), 
                    "AoA":"AoA_"+str(i+2)
                }
            ),
            how="outer",
            on="n"
        )
    return df_blades_aligned

def create_stack_plot_df(df_blades_aligned : pd.DataFrame) -> pd.DataFrame:
    """ This function creates a DataFrame that shows the consecutive
    difference between adjacent blades for each shaft revolution.
    
    Args:
        df_blades_aligned (pd.DataFrame): A DataFrame where every row contains the
        data pertaining to a single shaft revolution and every
        blade's ToA and AoA values are in its own column respectively. This
        is the output of the `pivot_blade_AoAs_along_revolutions` function.

    Returns:
        pd.DataFrame: A DataFrame where every row contains the
        data pertaining to a single shaft revolution and every
    """
    all_aoa_columns = sorted([
        i for i in df_blades_aligned.columns 
        if i.startswith("AoA_")
    ])
    B = len(all_aoa_columns)
    stack_plot_diffs = {}
    stack_plot_diffs["n"] = df_blades_aligned["n"].to_numpy()
    for blade_no in range(B - 1):
        farther_blade_name = all_aoa_columns[blade_no + 1]
        closer_blade_name = all_aoa_columns[blade_no]
        arr_blade_diffs = (
            df_blades_aligned[farther_blade_name] 
            - df_blades_aligned[closer_blade_name]
        ).to_numpy()
        
        stack_plot_diffs[closer_blade_name] = arr_blade_diffs
    farther_blade_name = all_aoa_columns[0]
    closer_blade_name = all_aoa_columns[B - 1]
    arr_blade_diffs = (
        df_blades_aligned[farther_blade_name].to_numpy()[1:] + 2*np.pi 
        - df_blades_aligned[closer_blade_name].to_numpy()[:-1]
    )
    arr_blade_diffs = np.append(arr_blade_diffs, [None])
    stack_plot_diffs[closer_blade_name] = arr_blade_diffs
    return pd.DataFrame(stack_plot_diffs)

def shift_AoA_column_headings(
    aoa_column_headings : List[str], 
    shift_by : int
) -> List[str]:
    """This function shifts the columns headings of the AoA
    such that the first column heading represents the first blade
    arriving at the first probe.

    Args:
        arr_values (np.ndarray): The array of values to be shifted
        shift_by (int): The number of positions to shift the values
        in the array by.

    Returns:
        List[str]: The shifted array of values.
    """
    if shift_by >= len(aoa_column_headings):
        raise ValueError(
            "shift_by must be less than the number blades in aoa_column_headings"
        )
    return (
        list(aoa_column_headings)[shift_by:] 
        + list(aoa_column_headings)[:shift_by]
    )

def rename_df_columns_for_alignment(
    df_to_align : pd.DataFrame,
    global_column_headings : List[str],
    shift_by : int
) -> pd.DataFrame:
    """This function performs two tasks. Firstly, it determines the mapping 
    between the global column headings and the column headings `df_to_align`. 
    Secondly, it renames and re-orders the columns in df_to_align such that 
    the columns appear in the same order in df_to_align.

    Args:
        df_to_align (pd.DataFrame): The DataFrame whose columns are to be
            renamed and re-ordered.
        global_column_headings (List[str]): The column headings
            to which the columns in df_to_align should be mapped. This
            will normally be AoA or ToA column headings.
        shift_by (int): The number of positions to shift the values
            in the array by.

    Returns:
        pd.DataFrame: The DataFrame with the renamed and re-ordered columns.
    """
    # Create a dictionary that maps the column headings in df_to_align
    # to the global column headings
    shifted_dataframe_columns = shift_AoA_column_headings(
        global_column_headings, 
        shift_by
    )
    column_headings_to_rename = {
        local_col : global_col
        for local_col, global_col 
        in zip(
            shifted_dataframe_columns,
            global_column_headings
        )
    }
    original_column_order = list(df_to_align.columns)
    df_to_align = df_to_align.rename(
        columns=column_headings_to_rename
    )
    return df_to_align[original_column_order]

def predict_probe_offset(
        df_probe_AoAs : pd.DataFrame,
        starting_aoa : float, 
        prox_probe_relative_distance : float,
    ) -> int:
    """This function calculates the offset that needs to be applied to
    the AoA columns of the current probe to align them with the first
    probe.

    Args:
        df_probe_AoAs (pd.DataFrame): A DataFrame where
            every row contains the data pertaining to a single shaft
            revolution and every blade's ToA and AoA values are in its
            own column respectively. This is the output of the
            `pivot_blade_AoAs_along_revolutions` function.
        starting_aoa (float): The mean AoA of the blade you want to
            project forward and identify in df_probe_AoAs. In radians.
        prox_probe_relative_distance (float): The relative distance
            between the current probe and the first probe. In radians.


    Returns:
        int: The blade offset that needs to be applied to the AoA values
            in df_probe_AoAs to align it to the blade in starting_aoa
    """
    predicted_blade_position = (
        starting_aoa 
        + prox_probe_relative_distance
    ) % (2*np.pi)
    all_aoa_columns = sorted([
        i for i in df_probe_AoAs.columns 
        if i.startswith("AoA_")
    ])
    current_probe_median_AoAs = df_probe_AoAs[all_aoa_columns].median()
    err_aoa = np.abs(current_probe_median_AoAs - predicted_blade_position)
    offset = np.argmin(err_aoa)
    return offset

def assemble_rotor_AoA_dfs(
    prox_aligned_dfs : List[pd.DataFrame], 
    probe_spacing : List[float]
) -> List[pd.DataFrame]:
    """This function assembles the rotor blade AoA DataFrames. In other
    words, this function receives the grouped AoA DataFrames from each
    probe, the one calculated by `pivot_blade_AoAs_along_revolutions` and 
    shifts the AoA values of each probe such that the first
    blade arriving at the first probe is aligned with the first blade
    arriving at the first probe. 

    We then assemble B DataFrames containing only the information
    from a single blade over every probe.

    Args:
        prox_aligned_dfs (List[pd.DataFrame]): A list of DataFrames
            where each DataFrame contains the ToAs and AoAs of a single
            blade from a proximity probe. Each DataFrame is the output
            of the `pivot_blade_AoAs_along_revolutions` function. 
        probe_spacing (List[int]): A list of relative probe spacing
            between the first probe and every other probe. There are one
            less value in this list than in prox_aligned_dfs.

    Returns:
        List[pd.DataFrame]: A list of DataFrames where each DataFrame
            contains the ToAs and AoAs of a single blade over all
            the proximity probes.
    """
    all_aoa_columns = sorted([
        i for i in prox_aligned_dfs[0].columns 
        if i.startswith("AoA_")
    ])
    all_toa_columns = sorted([
        i for i in prox_aligned_dfs[0].columns 
        if i.startswith("ToA_")
    ])
    remaining_columns = [
        i for i in prox_aligned_dfs[0].columns 
        if not i.startswith("ToA_") and not i.startswith("AoA_")
    ]
    B = len(all_aoa_columns)
    P = len(prox_aligned_dfs)
    if P  - 1 != len(probe_spacing):
        raise ValueError(
            "The number of proximity probes must be "
            "one less than the number of probe spacings"
        )
    rotor_blade_dfs = []
    for b in range(1, B+1):
        columns_to_copy = remaining_columns + [f"ToA_{b}", f"AoA_{b}"]
        rename_dict = {
            f"ToA_{b}" : "ToA_p1",
            f"AoA_{b}" : "AoA_p1"
        }
        rotor_blade_dfs.append(
            prox_aligned_dfs[0][columns_to_copy]
            .copy(deep=True)
            .rename(
                columns=rename_dict
            )
        )
    blade_1_probe_1_median = rotor_blade_dfs[0]["AoA_p1"].median()
    for iter_count, (df_probe_AoA, probe_offset) in enumerate(
            zip(prox_aligned_dfs[1:], probe_spacing)
        ):
        probe_no = iter_count + 2
        probe_offset = predict_probe_offset(
            df_probe_AoA,
            blade_1_probe_1_median,
            probe_offset
        )
        df_probe_AoAs_aligned = rename_df_columns_for_alignment(
            df_probe_AoA,
            all_aoa_columns,
            probe_offset
        )
        df_probe_AoAs_aligned = rename_df_columns_for_alignment(
            df_probe_AoAs_aligned,
            all_toa_columns,
            probe_offset
        )
        for b in range(1, B+1):
            columns_to_merge = ["n", f"ToA_{b}", f"AoA_{b}"]
            rename_dict = {
                f"ToA_{b}" : f"ToA_p{probe_no}",
                f"AoA_{b}" : f"AoA_p{probe_no}"
            }
            rotor_blade_dfs[b - 1] = rotor_blade_dfs[b - 1].merge(
                df_probe_AoAs_aligned[columns_to_merge].rename(
                    columns=rename_dict
                ),
                how="inner",
                on="n"
            )
    return rotor_blade_dfs




import numpy as np
import pandas as pd
from typing import List, Dict

from scipy.optimize import differential_evolution

def get_X(
        omega : np.ndarray,
        omega_n : float, 
        zeta: float,
        delta_st: float
    ) -> np.ndarray:
    """
    This function returns the vibration amplitude of 
    the blade vibration.
    
    x(ω) = 	δ_st / sqrt( (1 - r**2)**2 + (2*ζ*r)**2)

    where:

    r = ω/ω_0

    Args:
        omega (np.ndarray): The excitation frequencies in rad/s.
        omega_n (float): The natural frequency of the blade in rad/s.
        zeta (float): The damping ratio of the blade vibration.
        delta_st (float, optional): The static deflection of the blade. 
            This value is usually given in units of µm.

    Returns:
        np.ndarray: The amplitude of the blade vibration in the
            same units as delta_st.
    """
    r = omega / omega_n
    return (
        delta_st 
        / np.sqrt(
            (1 - r**2)**2 
            + (2*zeta*r)**2
        )
    )

def get_phi(
    omega : np.ndarray, 
    omega_n : float, 
    zeta: float
) -> np.ndarray:
    """Get the phase between the tip deflection and 
        the forcing function. 

    φ(ω) = arctan(2*ζ*r /  (1 - r**2))
    
    where:
    r = ω/ω_n

    Args:
        omega (np.ndarray): The excitation frequencies in rad/s.
        omega_0 (float): The natural frequency of the blade in rad/s.
        delta (float): The damping ratio of the blade vibration.

    Returns:
        np.ndarray: The phase of the blade vibration in rad.
    """
    r = omega / omega_n
    return np.arctan2(2 * zeta * r,1 - r**2)

def predict_sdof_samples(
    omega_n : float,
    zeta : float,
    delta_st : float,
    EO : int,
    theta_sensor : float,
    phi_0 : float,
    arr_omega : np.ndarray
) -> np.ndarray:
    """ This function determined the predicted SDoF fit
    samples at a proximity probe given the SDoF parameters.

    Args:
        omega_n (float): The natural frequency of the SDoF system.
        zeta (float): The damping ratio of the SDoF system.
        delta_st (float): The static deflection of the SDoF system.
        phi_0 (float): The phase offset of the SDoF system.
        EO (int): The EO of vibration you want to fit.
        theta_sensor (float): The sensor's angular position on the rotor.
        phi_0 (float): The phase offset of the SDoF system.
        arr_omega (np.ndarray): The angular velocity of the rotor corresponding
            to each revolution for which we want to predict the SDoF samples.

    Returns:
        np.ndarray: The predicted SDoF samples.
    """
    X = get_X(arr_omega*EO, omega_n, zeta, delta_st)  
    phi = get_phi(arr_omega*EO, omega_n, zeta)
    predicted_tip_deflections = X * np.cos(theta_sensor * EO - phi + phi_0)
    return predicted_tip_deflections

def get_correction_values(
    arr_omegas : float,
    z_median : float,
    z_max : float, 
) -> np.ndarray:
    """This function calculates the correction values for each sample
    based on the correction factors.

    Args:
        arr_omegas (float): The omega values for each sample.
        z_median (float): The correction value at the median shaft speed.
        z_max (float): The correction value at the max shaft speed.

    Returns:
        np.ndarray: The sample offsets for each sample.
    """
    omega_median = np.median(arr_omegas)
    omega_max = np.min(arr_omegas)
    m = (
        z_max
        - z_median
    ) / (
        omega_max 
        - omega_median
    )
    b = z_median - m * omega_median
    correction_values = m * arr_omegas  + b
    return correction_values

def SDoF_loss_multiple_probes(
        model_params : np.ndarray,
        tip_deflections_set : List[np.ndarray],
        arr_omega : np.ndarray, 
        EO : int, 
        theta_sensor_set : List[float],
        amplitude_scaling_factor : float = 1
    ) -> np.ndarray:
    """ This function fits the SDoF parameters to 
        multiple probes' data.

    Args:
        model_params (np.ndarray): The SDoF fit method's model parameters. It
            includes a list of the following parameters:
            
            omega_n (float): The natural frequency of the SDoF system.
            ln_zeta (float): The damping ratio of the SDoF system.
            delta_st (float): The static deflection of the SDoF system.
            phi_0 (float): The phase offset of the SDoF system.
            And then the z_median and z_max for each probe.
                z_median (float): The amplitude offset at the 
                    median shaft speed.
                z_max (float): The maximum amplitude offset.

        tip_deflections_set (List[np.ndarray]): The tip deflection data for each probe.
        arr_omega (np.ndarray): The angular velocity of the rotor corresponding
            to the tip deflection data.
        EO (int): The EO of vibration you want to fit.
        theta_sensor_set (List[float]): Each sensor's angular position 
            relative to the start of the revolution.
        amplitude_scaling_factor (float, optional): A scaling factor to
            weight the measured tip deflections. Defaults to 1. Use this value
            to reward solutions that better capture the full amplitude of the
            tip deflections.

    Returns:
        np.ndarray: The sum of squared error between 
            the tip deflection data of each probe and
            the predicted tip deflections.
    """
    omega_n, ln_zeta, delta_st, phi_0, *correction_factors = model_params
    zeta = np.exp(ln_zeta)
    error = 0
    for i_probe, arr_tip_deflections in enumerate(tip_deflections_set):    
        theta_sensor = theta_sensor_set[i_probe]
        predicted_tip_deflections = predict_sdof_samples(
            omega_n, zeta, delta_st, EO, theta_sensor, phi_0, arr_omega
        )
        z_median = correction_factors[i_probe*2]
        z_max = correction_factors[i_probe*2+1]
        arr_tip_deflection_corrections = get_correction_values(
            arr_omega, z_median, z_max
        )
        arr_tip_deflections_corrected = (
            arr_tip_deflections
            + arr_tip_deflection_corrections
        )
        error += np.sum(
            np.abs(arr_tip_deflections_corrected)**amplitude_scaling_factor
            *(
                arr_tip_deflections_corrected
                - predicted_tip_deflections
            )**2
        )
    return error

def perform_SDoF_fit(
    df_blade : pd.DataFrame,
    n_start : int,
    n_end : int,
    EOs : List[int] = np.arange(1, 20),
    delta_st_max : int = 10,
    verbose : bool = False
) -> Dict[str, float]:
    """This function receives a blade tip deflection DataFrame, and returns 
    the SDoF fit model parameters after fitting.

    Args:
        df_blade (pd.DataFrame): The blade tip deflection DataFrame.
        n_start (int): The starting revolution number of the resonance 
            you want to fit.
        n_end (int): The ending revolution number of the resonance 
            you want to fit.
        EOs (List[int], optional): The list of EOs to search for. Defaults 
            to np.arange(1, 20).
        delta_st_max (int, optional): The maximum static deflection within our optimization 
            bounds. Defaults to 10.
        verbose (bool, optional): Whether to print the progress. Defaults to False.

    Returns:
        Dict[str, float]: The fitted model parameters.
    """
    df_resonance_window = df_blade.query(f"n >= {n_start} and n <= {n_end}")
    measured_tip_deflection_signals = [
        col 
        for col in df_resonance_window
        if col.endswith("_filt")
    ]
    PROBE_COUNT = len(measured_tip_deflection_signals)
    eo_solutions = []
    for EO in EOs:
        if verbose:
            print("NOW SOLVING FOR EO = ", EO, " of ", EOs)
        omega_n_min = df_resonance_window["Omega"].min() * EO
        omega_n_max = df_resonance_window["Omega"].max() * EO
        ln_zeta_min = np.log(0.0001)
        ln_zeta_max = np.log(0.3)
        delta_st_min = 0
        phi_0_min = 0
        phi_0_max = 2*np.pi
        bounds = [
            (omega_n_min, omega_n_max),
            (ln_zeta_min, ln_zeta_max),
            (delta_st_min, delta_st_max),
            (phi_0_min, phi_0_max),
        ]
        tip_deflections_set = []
        theta_sensor_set = []
        for i_probe in range(PROBE_COUNT):
            z_max = df_resonance_window[f"x_p{i_probe+1}_filt"].abs().max()
            z_min = -z_max
            bounds.extend(
                [
                    (z_min, z_max),
                    (z_min, z_max)
                ]
            )
            tip_deflections_set.append(
                df_resonance_window[f"x_p{i_probe+1}_filt"].values
            )
            theta_sensor_set.append(
                df_resonance_window[f"AoA_p{i_probe+1}"].median()
            )
        multiple_probes_solution = differential_evolution(
            func = SDoF_loss_multiple_probes,
            bounds=bounds,
            args=(
                tip_deflections_set,
                df_resonance_window['Omega'].values,
                EO,
                theta_sensor_set,
                2
            ),
            seed=42
        )
        eo_solutions.append(multiple_probes_solution)
    best_EO_arg = np.argmin([solution.fun for solution in eo_solutions])
    best_EO = EOs[best_EO_arg]
    best_solution = eo_solutions[best_EO_arg]
    return {
        "omega_n" : best_solution.x[0] / (2*np.pi),
        "zeta" : np.exp(best_solution.x[1]),
        "delta_st" : best_solution.x[2],
        "phi_0" : best_solution.x[3],
        "EO" : best_EO,
    }


import pandas as pd
import numpy as np
from scipy.signal import butter, filtfilt

def get_blade_tip_deflections_from_AoAs(
    df_rotor_blade_AoAs : pd.DataFrame,
    blade_radius : float,
    poly_order : int = 11,
    filter_order : int = 2,
    filter_cutoff : float = 0.3
) -> pd.DataFrame:
    """This function performs the following operations:
        1. Normalizes the AoAs of each probe.
        2. Scales the AoAs to tip deflections.
        3. Detrends the tip deflections using a polynomial fit.
        4. Filters the tip deflections using a Butterworth filter.
        5. Calculates the peak-to-peak tip deflection.

    Args:
        df_rotor_blade_AoAs (pd.DataFrame): The DataFrame containing the AoAs of each 
            probe. This is an item from the list returned by the 
            function `get_rotor_blade_AoAs`.
        blade_radius (float): The radius of the blade in microns.
        poly_order (int, optional): The polynomial order to use for the detrending
            algorithm . Defaults to 11.
        filter_order (int, optional): The order of the butterworth filter. Defaults to 2.
        filter_cutoff (float, optional): The butterworth filter cutoff 
            frequency. Defaults to 0.3.

    Returns:
        pd.DataFrame: The DataFrame containing the detrended and filtered 
            tip deflections. This DataFrame also contains the peak-to-peak
            tip deflection.
    """
    df = df_rotor_blade_AoAs.copy(deep=True)
    all_aoa_columns = [
        col_name 
        for col_name 
        in df.columns 
        if col_name.startswith("AoA_p")
    ]
    butterworth_filter = butter(N=filter_order, Wn=filter_cutoff)
    for col in all_aoa_columns:
        df[col + "_norm"] = df[col].mean() - df[col]
        deflection_col_name = col.replace("AoA", "x")
        df[deflection_col_name] = blade_radius * df[col + "_norm"]
        poly = np.polyfit(df['Omega'], df[deflection_col_name], poly_order)
        df[deflection_col_name] = df[deflection_col_name] - np.polyval(poly, df['Omega'])
        df[deflection_col_name + '_filt'] = filtfilt(*butterworth_filter, df[deflection_col_name])
    x_matrix = df[[col for col in df.columns if col.endswith("_filt")]].to_numpy()
    df["pk-pk"] = x_matrix.max(axis=1) - x_matrix.min(axis=1)
    return df
